#!/usr/bin/env python
"""
Installing Local Resources (e.g. Databases and Tools)

Does install into directories without affecting the host setup

Uses
- Miniconda 
- Container image pulling

Maintains a .resources file within the project root for looking up install state from elsewhere

API:
    - python: E.g. pytest can run: get_redis_exe = lambda: project.get_present_resource_location('redis-server')
    - shell:  Use jq
"""


import os
from functools import partial

from devapp import app
from devapp.tools import FLG, project, exists, jsondiff, write_file, to_list

from . import _resources_api as api

app, run_app, do, system = app.app, app.run_app, app.do, app.system
g = api.g
inst_modes = ['', 'host', 'conda']

# resources = ['miniconda', 'redis-server', 'graph-easy']


class Flags(api.CommonFlags):
    autoshort = ''

    class force:
        n = '-y on all questions at install/uninstall'
        d = False

    class do_list:
        n = 'Output list of resources'
        d = True

    class do_install:
        n = 'Install comma seperated list of resources (match is enough)'
        d = []


S = api.S

# ------------------------------------------------------------------------ tools
present = lambda cmd: bool(os.popen('type "%s" 2>/dev/null' % cmd).read())


# ------------------------------------------------------------------------ actions


def install(match):
    rsc = api.matching_resource(match)
    return api.Install.resource(rsc)


# pkg_cmds = ['apt-get', 'dnf', 'yum']
# def pkgcmd():
#     c = S.pkg_cmd
#     if not c:
#         c = [k for k in pkg_cmds if present(k)]
#         c = c[0] if c else app.die('no package command', tried=pkg_cmds)
#     if c == 'dnf':
#         c += ' --color=false '
#     return c
# def install_host(rsc):
#     if rsc.installed:
#         return app.warn('already installed - skipping', rsc=rsc)
#     cmd = 'sudo %s install %s "%s" ' % (pkgcmd(), interactive(), rsc.pkg)
#     return do(system, cmd)


import json


def write_installed_resources(rscs):
    j = {}
    fn = project.fn_resources()

    # for rsc in rscs:
    #     j[rsc.name] = m = {'location': rsc.installed, 'present': True}
    #     v = api.g(rsc, 'post_inst_verify', 'x')
    #     if v != 'x':
    #         m['post_inst_verify'] = rsc.post_inst_verify
    r = [api.to_dict(i) for i in rscs if i.installed]
    write_file(fn, json.dumps(r, indent=2))
    app.warn('have written resources file', fn=fn)
    return j


# --------------------------------------------------------------------------- app


def run():
    rscs = api.find_resource_defs()
    if not FLG.do_install:
        FLG.do_list = True
    else:
        if not FLG.install_state:
            api.add_install_state(rscs)

    if FLG.do_install:
        [do(install, match=k) for k in FLG.do_install]
        if S.rsc_modified:
            api.add_install_state(rscs)
        return do(
            write_installed_resources, rscs=[c for c in rscs.listed() if c.installed]
        )

    if FLG.do_list:
        return api.to_list(rscs)


main = partial(run_app, run, flags=Flags)



def VimEnv(p):
    class Vim(p):
        runtime = 'user'
        filesystem = {
            'name': 'vimblack',
            'type': 'conda_env',
            'channel': 'https://my_company.com/pub/conda_axchange/channel',
            'packages': ['vim', 'vimblackhome'],
        }

    return Vim



"""
Process json logs. Default: Output to colorized ansi dev logs.

When lines cannot be json loaded we print them on stderr

"""
import sys, os, json, time

from devapp.tools import (
    skip_flag_defines,
    exists,
    to_list,
    repl_dollar_var_with_env_val,
    project,
    read_file,
    write_file,
)  # noqa: e402

skip_flag_defines.append('structlogging.sl')  # noqa: e402
from structlogging import sl
from devapp.app import FLG, app, run_app, do, system
from devapp import tools
from structlog import PrintLogger

levels = sl.log_levels


class Flags(sl.flags):
    autoshort = ''

    class file_name:
        n = 'file name of log file with json. "-": read from stdin'
        d = '-'

    class to_json:
        n = 'just print json loadable lines to stdout, rest to stderr. No ansi. Basically a filter, making jq work, when bogus lines are in.'
        d = False


def colorize(s, rend, p=PrintLogger(file=sys.stdout), levels=levels):
    try:
        s = json.loads(s)
        l = s['level']
        if levels[l] < app.log_level:
            return
        sys.stdout.write(rend(p, s['level'], s) + '\n')
    except:
        print(s, file=sys.stderr)


def run():
    fn = FLG.file_name
    if fn == '-':
        fd = sys.stdin
    else:
        if not exists(fn):
            app.die('not found', fn=fn)
        fd = open(fn)
    rend = sl.setup_logging(get_renderer=True)
    while True:
        s = fd.readline()
        if s:
            colorize(s, rend=rend)
            continue
        time.sleep(0.1)


main = lambda: run_app(run, flags=Flags)



import sys
from os import path

here = path.abspath(path.dirname(__file__))


def is_installed_in(da_dir_max):
    # DEPRECATED
    """Called at max init with its location derived $DA_DIR in order to see
    if its python libs are installed correctly
    No need to say anything, most errors there won't even reach this point,
    covers only the case when an other version of the libs are active.
    """
    if here.startswith(da_dir_max.rsplit('/', 1)[0]):
        return
    sys.exit(1)



"""
Templating

"""
import json
import os
import re
import sys

from devapp import tools
from devapp.app import app
from tree_builder import full_id

# passed into rendering contexts:
r_objs_by_fid = {}

nil = '\x01'
fidstr = lambda r: r['___']  # stored here to show up top

postprocs = {'': lambda v: v, 'json.dumps': json.dumps}


class SmartAdapt:
    """processing the context render hits on cls.<key>"""

    def __init__(self, r, name=''):
        name = 'Template ' + name
        self.r, self.name = r, name
        self.log = app.log.bind(name=name, fid=fidstr(r))

    def __getitem__(self, k):
        'try proxy to class first'
        k, postproc = (k + '|>').split('|>')[:2]
        v = self.getitem(k)
        return postprocs[postproc](v)

    def getitem(self, k):
        if k == 'fid':  # often
            return self.r['spec']['fid']
        r, m = self.r, self.r
        k, dflt = (k + '|' + nil).split('|')[:2]
        ko = k
        if not '.' in k:
            if k.startswith('v:'):
                return Vault.get(r, k[2:])
            k = 'env.' + k
        elif k.startswith('.'):
            k = k[1:]
        parts = k.split('.')
        k = parts.pop()
        for part in parts:
            try:
                m = m[part]
            except:
                raise Exception(self.name, 'part not present', part, 'in', ko)
        # check env, alternatively:
        v = m.get(k, r['env'].get(ko, dflt))
        if v == nil:
            if sys.stdin.isatty():
                app.warn('Replacement Error', key=ko, r=r)
                self.log.warn('SmartAdapt: could not resolve', key=ko)
                breakpoint()
                print('Inspect the environ here')
            app.die('Could not resolve', key=ko, r=r)
        return v


def execfn(fn, r, smart=None):
    out = []
    # fmt:off
    # The rendering globals:
    m = {
        'r'        : r,
        'api'      : API,
        'R'        : smart,
        'out'      : out,
        # 'Vault'  : Vault,
        'classes'  : r_objs_by_fid,
    }
    # fmt:on
    # if template chdirs
    here = os.getcwd()
    try:
        # chdir to the hook file, to make live easy there:
        os.chdir(os.path.dirname(fn))
        app.info('Executing as python', template=fn)
        # better than exec when debugging:
        tools.exec_file(fn, m)
    finally:
        os.chdir(here)
    return out


single_percent_regex = re.compile(r'%(?!\()')  # perc. not followed by bracket


def exec_templ(fn, r):
    """we want to keep the name fixed, no matter if its plain or code
    (better for debugging)
    So we require a tag "axc-exec" somewhere within to exec it
    - execs are always execed using the current ctx
    - plains(simples) are loaded once, then mapped n times.
    """
    if not tools.exists(fn):
        raise Exception(fn, 'does not exist')
    smart = SmartAdapt(r, fn)
    out = tools.read_file(fn, '', bytes=100)
    first_line = out.split('\n', 1)[0]
    if first_line.startswith('#!') and 'python' in first_line:
        # if the template is python its result is a line array, 'out'
        out = '\n'.join(execfn(fn, r, smart))
    else:
        app.info('Reading', template=fn)
        out = tools.read_file(fn)
    try:
        try:
            return out % smart
        except TypeError as ex:
            pr = single_percent_regex.sub('%%', out)
            if pr != out:
                _ = 'You have single "%" in your template. '
                app.warn(_ + 'Trying with "%" escaped version.')
                return pr % smart
    except Exception as ex:
        app.warn(ex)
        breakpoint()
        _ = 0  # just to stay in local scope for ex


class API:
    """
    Template API.
    TODO:Global "api" in hooks.
    For reliabilty compared to uncontrollable
    imports of anything, this we can guarantee stable.
    => Never remove stuff here!'
    """

    class links:
        def get_via(r, io, *caps, min=1, max=1000):
            """
            conn = lambda io, w: api.links.get_via(r, io, w, min=1)[0]
            httpapi = conn('in', 'api')
            """
            from devapp.spec.links import get_via_by_r

            return get_via_by_r(r, io, *caps, min=1, max=1000)

        def get_vias(links):
            """for simple svcs easier to handle"""
            all_vias, add = {}, False
            for d in 'in', 'out':
                vias = all_vias[d] = []
                conns = links[d]
                for c in conns:
                    v = {'cls': ['.'.join(full_id(co)) for co in c['cls']]}
                    v.update(c['via'])
                    vias.append(v)
                    add = True
            return all_vias

    class inventory:
        pass

    class opsys:
        pass  # don't collide with import os after _into

    class fs:
        pass

    @staticmethod
    def create_token(r, secret_at=None, **kw):
        """TODO: support central token stores, Vault
        secret at is ignored, too shaky with spec dependent hub names
        """
        import jwt, time, uuid

        # independent of hub name:
        s = r['env']['DA_DIR'] + '/secure/token_secret'
        fn = r['env'].get('DA_FILE_TOKEN_SECRET', s)
        if not os.path.exists(fn):
            app.warn('Creating token secret', secret_file=fn)
            tools.write_file(fn, str(uuid.uuid4()))
        s = tools.read_file(fn)

        kw['iat'] = int(time.time())
        kw['exp'] = kw.get('exp', kw['iat'] + 86400 * 360 * 5)
        kw['why'] = kw.get('why', 'build')
        return jwt.encode(kw, s, algorithm='HS256').decode('utf-8')

    # links = Links
    @staticmethod
    def lazy(*a, **kw):
        'provide a cache for user funcs, e.g. the ssh key read'
        # not overwrite our own stuff:
        # TODO: take from old axc2:
        return c_lazy('user_api_' + a[0], *a[1:], **kw)

    @classmethod
    def _into(cls, g):
        """convenience, hooks *can* call this to have api funcs in globals
        api.write_file -> write_file..."""
        [
            operator.setitem(g, k, getattr(cls, k))
            for k in dir(cls)
            if not k.startswith('_')
        ]


'These we pass into custom funcs to make live simpler there'
# we do these one by one not by module
# to avoid accidential changes of function names in modules, breaking hooks

# api_funcs = (full_id, say, warn, l, die, call_func, dict_to_txt)
# [setattr(API, k.__name__, staticmethod(k)) for k in api_funcs]

# opsys api:
# os_api_funcs = (app.sh,)
# [setattr(API.opsys, k.__name__, staticmethod(k)) for k in os_api_funcs]

# fs api, pub funcs marked with 'API:' there:
# for k in dir(fs):
#    f = getattr(fs, k)
#    if (getattr(f, '__doc__', 0) or '').startswith('API:'):
#        setattr(API.fs, k, staticmethod(f))


def align_by_eq(lines):
    """make a nice aligned unit file """
    max_len, out, kvs = 0, [], []

    def add_kvs(out, kvs, m=0):
        if not kvs:
            return
        kls = [len(kv[0]) for kv in kvs if isinstance(kv, tuple)]
        if kls:
            m = max(kls)
        ind = ' ' * (m + 3)
        for kv in kvs:
            if not isinstance(kv, tuple):
                out.append(ind + kv)
            else:
                out.append('%s = %s' % (kv[0].ljust(m), kv[1]))

    while lines:
        line = lines.pop(0)
        if not line:
            kvs.append(line)
        elif line[0] == '[':
            add_kvs(out, kvs)
            out.append(line)
            kvs = []
        else:
            kv = line.split('=', 1)
            if len(kv) != 2:
                kvs.append(line)
            else:
                kvs.append((kv[0].rstrip(), kv[1].lstrip()))
                while lines:
                    if not line.endswith('\\'):
                        break
                    line = lines.pop(0)
                    kvs.append(line)
    if kvs:
        add_kvs(out, kvs)
    return '\n'.join(out)



#!/usr/bin/env python
"""
source env.sh
python -m devapp.run start

"""
import importlib
import json
import os
import subprocess as sp
import sys
import time

import psutil
from devapp.spec import tools as spec_tools
from devapp.tools import read_file, restart_unshared, write_file

[sys.path.append(p) for p in os.environ.get('sys_path_append', '').split(':')]


env = os.environ
exists = os.path.exists
FS = spec_tools.FS
Proc = psutil.Process


def out(msg, kw):
    print(
        msg, ', '.join(['%s=%s' % (k, str(v)) for k, v in kw.items()]), file=sys.stderr,
    )


# we don't have logging before we have the fileystem(flags)
def info(msg, **kw):
    out('Info: %s' % msg, kw)


def warn(msg, **kw):
    out('Warning: %s' % msg, kw)


def dbg(msg, **kw):
    out('Dbg : %s' % msg, kw)


def die(msg, **kw):
    out('Fatal: %s' % msg, kw)
    sys.exit(1)


class InitApp:
    """mocks devapp.app while we not yet have it on app initializing"""


InitApp.die = die
InitApp.dbg = InitApp.debug = dbg
InitApp.info = info


def verify_env_sourced():
    try:
        db = FS.d_build()
        fn_env = db + '/env.sh'
    except:
        die('Please source an env.sh file before running')
    if not exists(fn_env):
        die('App is not built yet, missing file', env_file=fn_env)
    # cur = env.get('build_dir', '')

    # if cur and cur == db:
    #    # already sourced via source env.sh (e.g. systemd - unit):
    #    return
    # from devapp.spec import os_tools

    # env.update(os_tools.source(fn_env, set_sys_path_from_py_path=True))


now = lambda: int(time.time() * 1000)
is_svc = lambda: env['type'] == 'Service'
dirname = os.path.dirname
pidfile = lambda: env['var_dir'] + '/pidfile'
pidmain = lambda: env['var_dir'] + '/pidfile.main'
unit_name = lambda: env.get('name_unit')


def from_pidfile():
    i = int(read_file(pidfile(), -1))
    return i if psutil.pid_exists(i) else 0


# default if user provided nothing else:
bash = lambda: ['bash', '--rcfile', '%(DA_DIR_DEVAPPS)s/shell/bashrc' % env]


def write_pid(pid=None):
    # is given for spawned ones else its ours:
    pid = pid or os.getpid()
    # for devapps both are ident:
    for fn, p in (pidfile(), pid), (pidmain(), LifeCycle.main_pid):
        write_file(fn, str(p) + '\n', mkdir=True)


def run_exit(args):
    # TODO: basic support for && and pipes
    args = [str(l) for l in args]
    proc = sp.Popen(args)
    write_pid(proc.pid)
    proc.communicate()
    sys.exit(proc.returncode)


def js_to_struct(s, dflt=list):
    r = dflt() if not s else s
    return json.loads(r) if isinstance(r, str) else r


def is_in_run_dir(mod):
    dr = env['run_dir']
    if exists(dr + '/%s.py' % mod):
        sys.path.insert(0, dr) if not dr in sys.path else 0
        return True


class LifeCycle:

    main_pid = None

    def run(argv_cmd=None):
        """argv_cmd != None if user just wants to enter filesystem
        If so and servie is nspawn -> we nspawn without boot.
        otherwise we run his cmd (default bash)
        """
        breakpoint()
        LifeCycle.main_pid = os.getpid()
        env['DA_INITTED'] = ''  # $PATH exports done in init

        FS.initapp = InitApp
        FS.build_from_fs_stack()

        appl, app_mod = [env.get('app')], None
        if (not '/' in appl[0] and '.' in appl[0]) or is_in_run_dir(appl[0]):
            try:
                # exists not -> must be devapp module:
                app_mod = importlib.import_module(appl[0])
            except Exception as ex:
                warn('Cannot import app_module', mod=appl[0], exc=ex)

                # die('Cannot import %s:' % appl[0], error=ex)

        app_args = js_to_struct(env.get('app_args'), list)
        # the pid we keep is not ours and also not that of the inner init
        # but that of nspawn, which propagates graceful termination of inner services:
        if argv_cmd:
            if appl == ['systemd-nspawn']:
                for b in '--boot', '-b':
                    app_args.remove(b) if b in app_args else ''
                argv_cmd = appl + app_args + argv_cmd
            run_exit(argv_cmd)

        if not app_mod:
            run_exit(appl + app_args)

        # ok - inline (r/f)un:
        write_pid()  # yup, thats us :-)
        flagfile, flags = '', env.get('DA_FILE_FLAGS', 'prod')
        if '/' in flags:
            flagfile = flags
            if not exists(flagfile):
                die('Not found', flagfile=flagfile)
        else:
            for d in 'conf_dir', 'etc_dir':
                fn = env.get(d) + '/%s.flags' % flags
                if exists(fn):
                    flagfile = fn
                    break
        if flagfile:
            # The implicit find-by-presence of flagfiles is nice but absl-py
            # must see the --flagile argument. Not PC but we have to tune the
            # sys.argv for this. Oh, and being at it, lets streamline $0 as
            # well:
            # DA_CLS results in app.name which determines the folder where
            # secrets are searched, e.g. hub -> LEAVE!:
            sys.argv.insert(1, flagfile)
            sys.argv.insert(1, '--flagfile')
            # sys.argv.insert(0, env['DA_CLS'])
        info('DevApp', app=appl, args=sys.argv)
        # sys.path.append('/data/f3/repos/axess-4/streampipes.git/src')
        app_mod.run()

    def procs():
        """
        Main: This is us - the python program which runs inline - OR spawns the
        child process, like nspawn
        Enter: The proc for nsenter. For nspawn:boot the grand child, i.e. init
        = Main for devapps.
        Kill: This is the one in the pid file.
        The pid which gracefully terminates all childs, i.e. nspawn for those
        (not init).  Main for devapps.
        """

        pkill, pmain = read_file(pidfile(), 0), read_file(pidmain(), 0)
        # don't trust pidfiles:
        try:
            PK, PM = Proc(int(pkill)), Proc(int(pmain))
            # proc is running
            rt = env.get('runtime', '')
            PE = PK
            if rt == 'systemd:boot':
                PE = PK.children()[0]
        except:
            PM, PK, PE = None, None, None
        return {'proc_main': PM, 'proc_kill': PK, 'proc_enter': PE}

    def is_systemd(procs=None):
        P = (procs if procs else LifeCycle.procs())['proc_main']
        return None if not P else True if P.parent().pid == 1 else False

    is_running = lambda: bool(LifeCycle.procs()['proc_kill'])

    def pids():
        m = {}
        for k, P in LifeCycle.procs().items():
            m[k.replace('proc_', 'pid_')] = P.pid if P else 0
        return m

    def enter():
        ep = LifeCycle.pids()['pid_enter']
        if not ep:
            die('Not running', process=env['DA_CLS'])
        env['DA_INITTED'] = ''  # causes init. (for PATH). nsenter keeps env
        c = ('nsenter -t %s -m -u -i -n -p --' % ep).split(' ')
        c.extend(argv_cmd())
        sys.exit(sp.call(c))

    def status():
        """ for linux """
        LC, pids = LifeCycle, dict(LifeCycle.pids())
        pids['unit'], runs = unit_name(), pids['pid_main']
        m, msg = (app.info, 'Is Running') if runs else (app.warn, 'Is stopped')
        m(msg, **pids)
        sys.exit(0 if runs else 1)

    def show():
        """Svc started by systemd OR manually - then we check pidfile"""
        st = ['systemctl', 'show', unit_name(), '--no-page']
        out, err = sp.Popen(st, stdout=sp.PIPE, stderr=sp.PIPE).communicate()
        out = out.decode('utf-8').splitlines()
        d = dict([(k, v) for k, v in [l.split('=', 1) for l in out]])
        d.update(LifeCycle.pids())
        return d

    def stop():
        # nginx needs special stopping:
        stp = js_to_struct(env.get('service_exec_stop', []))
        if stp:
            sys.exit(sp.call(stp))
        procs = LifeCycle.procs()
        if LifeCycle.is_systemd(procs):
            # he can do it better than us. Plus we don't want auto-restarts:
            sys.exit(sp.call(['systemctl', 'stop', unit_name()]))
        if not procs['proc_kill']:
            return info('Not running')
        os.kill(procs['proc_kill'].pid, 13)

    def fs(*args):
        """just build the fs of a service"""
        LifeCycle.start(argv_cmd())

    def not_impl():
        raise Exception('Not Supported')

    def pstree(depth=2):
        procs = LifeCycle.procs()
        if not procs['proc_kill']:
            return {}

        def walk(P):
            tree = {}
            tree[P.pid] = m = {'proc': P}
            childs = P.children()
            if childs:
                m['tree'] = mc = []
                for c in childs:
                    mc.append(walk(c))
            return tree

        return walk(procs['proc_main'])

    def systemctl(*args):
        """Adding the unit name to systemctl when it makes sense"""
        if not args:
            args = ('status',)
        args = list(args)
        if args[0] == 'status' or (
            len(args) == 1 and not args[0][:4] in (['list', 'daem'])
        ):
            args.append(unit_name())
        if os.geteuid():
            args.insert(0, '--user')
        app.info(' '.join(['/usr/bin/systemctl', *args]))
        sys.exit(sp.call(['/usr/bin/systemctl', *args]))

    sc = systemctl  # shortcut


argv_cmd = lambda: sys.argv[2:] or bash()


def _main():
    verify_env_sourced()
    m = getattr(LifeCycle, sys.argv[1], LifeCycle.not_impl)
    res = m(*sys.argv[2:])
    return res


def main():
    a = sys.argv
    # auto on no args:
    if len(a) == 1 or not getattr(LifeCycle, a[1], None) or a[1] == 'main':
        # main is auto: enter or fs, depending on running or not
        # if LifeCycle.is_running():
        #    a.insert(1, 'enter')
        # else:
        # Changed that to always just enter filesystem (nsenter requires root):
        # also when he entered before the start it would be confusingly different
        a.insert(1, 'fs')

    if sys.argv[1] in ('start', 'fs'):
        # we assume the user who builds is the user who starts. If root we allow unshare requiring fs components
        # for user we are not there yet, examining bubblewrap or podman user mode... but not yet supported
        # I.e. we only need to unshare when we are root right now:
        if not os.geteuid():
            # if not int(os.getppid()) == 1 and (not os.geteuid() or 1):
            # if systemd starts us we took care that the unit file has unshare in the
            # ExecStart - IF needed. For normal users we unshare when he goes into
            # daemon FS:
            # If normal user does absolutely not want unshared entries he can export
            # da_unshared=$DA_CLS (see this restart_unshared func):
            restart_unshared(env['DA_CLS'])

        # not parse flags yet if this is start:
        if sys.argv[1] in ('start',):
            sys.exit(_main())
    # ok this is not a start action where devapp.app must be the target app
    # (there can be ony one, alone for flags parsing)
    # So, WE are the one app:
    global app
    from devapp.app import app, run_app

    run_app(_main, flags_parser=extract_app_flags)


def extract_app_flags(args):
    """
    We solve the tricky problem that we want for non start runs, i.e. where
    WE are the devapp the parsing of standard flags like --log_level - but also
    not crash for "flag-style" args like -lta in `fs --log-level=10 ls -lta`
    The solution is this custom parser, which stops at first non '-' arg when
    handing over to FLAGS
    """
    from absl import flags

    a, meth_args = args[:2], []
    into = a
    for p in args[2:]:
        if not p.startswith('-'):
            into = meth_args
        into.append(p)
    res = flags.FLAGS(a)
    sys.argv.clear()
    sys.argv.extend([*res[:2], *meth_args])
    return sys.argv


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Service Lifecycle Operations

Only for services where systemd units had been created
"""
# Could be done far smaller.
from re import match
from devapp import gevent_patched
import hashlib
import json
import os
import shutil
from copy import deepcopy
from functools import partial
import sys

import requests
from devapp.app import FLG, app, run_app, do, system
from devapp import tools
from devapp.tools import (
    exists,
    to_list,
    repl_dollar_var_with_env_val,
    project,
    read_file,
    write_file,
)
import devapp.resource_tools as api


class Flags(api.CommonFlags):
    autoshort = ''

    class match:
        n = 'Restrict to matching units'
        d = ''

    class stop_all:
        n = 'When given we kill ALL processes occupying the port of a service'
        d = False


S = api.S


def run():
    argv = sys.argv[2:]
    if FLG.match:
        argv = argv[2:]

    d = project.root()
    db = d + '/bin'
    if not exists(db):
        app.die('Not in project dir with bin directory', required=db)
    wrappers = {}
    units = []
    for u in os.listdir(db):
        s = wrappers[u] = read_file(db + '/' + u, dflt='')
        if 'systemctl ' in s and FLG.match in u:
            units.append(u)
    if not units:
        msg = 'No units - we can only manage services where unit files had been created'
        app.die(msg)
    app.info('units', units=units)
    if FLG.stop_all:
        argv = ['stop']
        app.warn('Stop all is set - we stop everything at unit ports')
    else:
        if not argv:
            argv = ['status']

    cmd = ' '.join(argv)
    for u in units:
        if cmd == 's':
            cmd = 'start'
        app.info(cmd + ' ' + u)
        do(system, db + '/' + u + ' ' + cmd, no_fail=True)

    if cmd == 'start':
        show_failing(units, db)
    elif FLG.stop_all:
        [stop_all(unit, wrappers[unit]) for unit in units]


def get_unit(pid):
    u = os.popen('systemctl --user status %s | head -n 1 | cut -d " " -f 2' % pid)
    return u.read().strip()


def stop_all(unit, wrapper):
    port = wrapper.split('port=', 1)
    if not len(port) - 1:
        return
    port = int(port[1].split('\n', 1)[0].replace('"', '').replace("'", ''))
    pids = os.popen('lsof -ti tcp:%s' % port).read().strip().splitlines()
    if not pids:
        return
    app.info('killing other %s at port %s' % (unit, port), pids=pids)
    for p in pids:
        unit = get_unit(p)
        if unit:
            do(system, 'systemctl --user stop "%s"' % unit)
            continue
        os.kill(int(p), 15)
        app.info('killed', pid=p)


def show_failing(units, db):
    errs = []
    for u in units:
        err = os.system(db + '/' + u + ' status >/dev/null')
        if err:
            os.system(db + '/' + u + ' status')
            errs.append(u)
    if errs:
        app.die('Could not start', services=errs)


main = lambda: run_app(run, flags=Flags)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Ansi Escape Tools for Terminal Output

"""
from __future__ import absolute_import, division, print_function

import os
import sys
import time
from collections import OrderedDict
from functools import partial

from theming.unicode_chars import blocks

# https://misc.flogisoft.com/bash/tip_colors_and_formatting


if sys.version_info[0] > 2:
    char = chr
else:
    char = lambda nr: unichr(nr).encode('utf-8')

envget = os.environ.get


RESET = '\x1b[0m'
BRIGHT = '\x1b[1m'
DIM = '\x1b[2m'


class Theme:
    """
    Basic *Semantic* Highlighting Functions.
    Because it makes no sense to hardcode specific colors but the *meaning* of
    something to be shown.

    Offering Appropriate Coloring of Output for:
    - I: Important
    - M: Marked Up
    - L: Low Intensity
    - D: Dimmed Away
    - R: Red (Error)
    - G: Green (OK)

    The color codes are taken from the environ if defined, allowing for
    consistent theming with other tools.
    E.g. echo -e "${M}Big$R Error" equals
         print(M('Big'), R('Error'))

    We use 256 colors - if you have an 8 color term, then detect it on the
    # parent process and overrite these by exporting e.g. `export R="\\e[1;34m"`
    """

    # fmt:off
    I = [11, 'Important', 'Titles, H1, New contexts']
    G = [2, 'Green', 'OK, H2, Success']
    M = [1, 'Marked UP', 'Sub Titles, H3, Important values in dicts']
    R = [124, 'Red', 'Alarm, Error, Warning, Failed']
    L = [5, 'Low Intensity', 'Debug Output, Code, Shell output']
    D = [5, 'Dimm Away', 'Very Unimportant, No Distraction']
    bold = 'MIR'
    dimm = 'D'
    ignore_env = False
    force_colors = False  # even if no tty
    disabled = False  # no coloring
    show_order = 'IGMRLD'
    name = 'Default'
    # fmt:on
    #    colorful
    #    fruity
    #    monokai
    #    paraiso_light
    #    rrt
    #    trac
    #    abap
    #    autumn
    #    default
    #    igor
    #    murphy
    #    pastie
    #    sas
    #    vim
    #    algol
    #    borland
    #    emacs
    #    lovelace
    #    native
    #    perldoc
    #    stata
    #    vs
    #    algol_nu: cool
    #    bw
    #    friendly
    #    manni
    #    paraiso_dark
    #    rainbow_dash
    #    paraiso-dark : good
    pygments_style = 'paraiso-dark'

    # can't derive the termwidth in process, this is expensive:
    default_width = 80
    _cur_width = None
    width_check_secs = 2
    _cur_width_last_check = 0
    _cur_width_check_failing = False

    @classmethod
    def output_shell_exports():
        """prints a sourceable set of export statements so that a theme
        defined in python can be sourced
        (i.e. reversing the definition order, we normally use the shell theme)
        """
        raise NotImplemented

    @classmethod
    def attrs(cls):
        """Other attributes of the Theme"""
        c = cls.color_keys()
        r = {'name': cls.__name__}
        for k in dir(cls):
            if k.startswith('_'):
                continue
            v = getattr(cls, k)
            if isinstance(v, (str, int, bool, float)):
                r[k] = v
        return r

    @classmethod
    def width(cls):
        if cls._cur_width is None and cls._cur_width_last_check > 0:
            return cls.default_width
        if cls._cur_width_check_failing:
            # might have been provided newly with new theme:
            # -> ignore current
            return cls.default_width
        now = time.time()
        if now - cls._cur_width_last_check < cls.width_check_secs:
            return cls._cur_width or cls.default_width
        c = os.popen('tput cols 2>/dev/null').read()
        cls._cur_width_last_check = now
        if not c:
            # thats it, we won't check again:
            cls._cur_width_check_failing = True
            return cls.default_width
        c = cls._cur_width = int(c)
        return c

    @classmethod
    def line(cls, col='L'):
        cols = cls.width()
        return cls.colorize(col)('_' * cols)

    @classmethod
    def colorize(cls, col):
        return getattr(cls, col.lower())

    @classmethod
    def color_keys(cls):
        r = []
        s = cls.show_order
        while s:
            c, s = s[0], s[1:]
            r.append(c) if hasattr(cls, c) else 0
        [r.append(c) for c in dir(cls) if not c in r and cls.is_col_key(c)]
        return r

    @classmethod
    def colorizers(cls):
        """ordered list of colorizer functions"""
        return [cls.colorize(k) for k in cls.color_keys()]

    @staticmethod
    def is_col_key(k):
        return len(k) < 3 and k == k.upper()

    @classmethod
    def setup(cls, **alt_theme):
        """
        We allow to provide new color shortcuts as well but they have to obey
        the convention to be < 3 characters and upper case
        Color names and tags can be given by providing a list and not the code.
        E.g. setup(I=123, X=[333, 'My Title', 'Foo, Bar'])
        """

        def col_def(k, v, dt='Unnamed %s Color', tt='<Untagged>'):
            if not Theme.is_col_key(k):
                return v
            if isinstance(v, int):
                return [v, dt % v, tt]
            # add not given infos:
            v.append(0) if len(v) == 0 else 0
            v.append(dt % v[0]) if len(v) == 1 else 0
            v.append(tt) if len(v) == 2 else 0
            return v[:3]

        [setattr(cls, k, col_def(k, v)) for k, v in alt_theme.items()]
        cols = cls.color_keys()
        [setattr(cls, c.lower(), cls._colorizer(c)) for c in cols]

    @classmethod
    def _colorizer(cls, key, have_tty=sys.stdout.isatty()):
        """returns the callable which colorizes a string"""
        do_cols = have_tty or cls.force_colors
        if cls.disabled:
            do_cols = False

        dflt = getattr(cls, key, [''])[0]

        def colorize(s, into):
            return into % s

        # bold / dimm
        b = '1;' if key in cls.bold else '2;' if key in cls.dimm else ''

        e, t = '', '%s'
        if do_cols:
            if not cls.ignore_env:
                # TODO perf?:
                e = envget(key, '')
                if e.isdigit():
                    e = '1;38;5;%sm' % e
                elif e:
                    e = e.split('[', 1)[1]
            if not e and dflt:
                e = b + '38;5;%sm' % dflt
            if e:
                # turns out to be the safest
                t = ''.join(('\x1b[', e, '%s', RESET))

        return partial(colorize, into=t)

    @classmethod
    def color_prefixes(cls):
        """returns just the prefix w/o the reset at the end"""
        d = lambda k: getattr(cls, k.lower())('XX').split('XX')[0]
        return dict([(k, d(k)) for k in cls.color_keys()])

    @classmethod
    def definition(cls, key):
        d = getattr(cls, key)
        m = {
            'title': d[1],
            'code': d[0],
            'tags': [i.strip() for i in d[2].split(',')],
            'func': getattr(cls, key.lower()),
        }
        return m

    @classmethod
    def print(cls):
        """
        Print the Theme.
        #TODO Not very convincing at the moment, make a bit fancier
        """
        p = print
        kw = cls.attrs()
        p(cls.line('I'))
        p(I('Theme %s' % kw.pop('name')))
        p(cls.line('I'))
        [p('- %s: %s' % (k.ljust(20), M(v))) for k, v in kw.items()]

        p(cls.line())

        for c in cls.color_keys():
            d = cls.definition(c)
            f = d['func']
            p(I(c), '[%s]' % d['code'], d['title'])
            p(cls.line())
            [p('- %s' % f(i)) for i in d['tags']]
            p('')
        [
            p(*l)
            for l in (
                ('',),
                (cls.line('I'),),
                ('Testoutput',),
                ('I: ', I('Title')),
                ('M: ', M('Sub Title')),
                ('R: ', 'This is an ', R('Error')),
                ('G: ', 'But %s worked well (G)' % G('that')),
                ('L: ', L('[debug]'), 'output.....'),
                ('D: ', D('Totally uninteresting')),
            )
        ]


Theme.setup()
# Its colorizer functions are now set - e.g. Theme.r('error')
# they HAVE to be supershort to not distract the reader of code:
M = Theme.m
I = Theme.i
L = Theme.l
R = Theme.r
G = Theme.g
D = Theme.d


# This is to format a column e.g. with Thread Number
# distinctively:

_bnames = 'arrows', 'geometric_shapes', 'box_drawing'
_chars = list(range(48, 232))  # number, letters
[_chars.extend(list(getattr(blocks, k))) for k in _bnames]
# we go down the colums here
# https://misc.flogisoft.com/bash/tip_colors_and_formatting
# to get visually differing colors. hsl would be better
alternating_colors = [col * 10 + row for row in range(0, 10) for col in range(0, 25)]


class Cell:
    """formats one cell"""

    @staticmethod
    def unique(nr, _lchars=len(_chars), _lcols=len(alternating_colors) - 1):
        """
        Returns a visually unique cell in the terminal
        combining bg and fg colors plus symbols in black

        The passed nr argument must be int, and, well, unique for the same
        output
        """
        try:
            symb = char(_chars[nr % _lchars])
        except:
            symb, nr = nr, hash(nr)
        return Cell.colorize(symb, nr)

    @staticmethod
    def colorize(symb, nr, _lcols=len(alternating_colors) - 1):
        bg = alternating_colors[nr % _lcols]
        return '\x1b[1;38;5;201;2;48;5;%sm%s%s' % (bg, symb, RESET)


def unique(s):
    return Cell.unique(s)


def unique_str(s):
    return Cell.colorize(s, nr=hash(s))


def structlog_style():
    """for structlog init w/o colorama (e.g. ax.rx)"""
    #  structlog/dev.py
    C = Theme.color_prefixes()

    # fmt:off
    def style(orig_style):
        class S(orig_style):
            level_critical = level_exception = level_error = C['R']
            level_warn  = C['I']
            level_info  = C['M']
            level_debug = C['L']
            timestamp   = C['D']
            logger_name = C['L']
            kv_key      = C['M']
            kv_value    = C['I']
            reset       = RESET

        return S

    ls = {
        'critical'  : C['R'],
        'exception' : C['R'],
        'error'     : C['R'],
        'warn'      : C['I'],
        'warning'   : C['R'],
        'info'      : C['G'],
        'debug'     : C['L'],
        'notset'    : '',
    }
    # fmt:on

    return style, ls


# old or new version? we keep or now
# def structlog_style(colors=True):
#    '''for structlog init w/o colorama (e.g. ax.rx)'''
#    #  structlog/dev.py
#    C = Theme.color_prefixes()
#    # fmt:off
#    ls = {
#             "critical"  : C['R']
#           , "exception" : C['R']
#           , "error"     : C['R']
#           , "warn"      : C['I']
#           , "warning"   : C['R']
#           , "info"      : C['G']
#           , "debug"     : C['L']
#           }
#    ks = {
#             "timestamp" : C['D']
#           , "kv_key"    : C['L']
#           , "kv_value"  : C['M']
#           , "logger"    : C['M']
#           , "meta"      : C['M']
#           , "event"     : ''
#           , "notset"    : ''
#           }
#    # fmt:on
#    if not colors:
#        ls = dict((k, '') for k in ls)
#        ks = dict((k, '') for k in ks)
#
#    longest_level = len(max(ls, key=lambda l: len(l)))
#    justed_levels = dict((k, k.ljust(longest_level)) for k in ls)
#    return {
#        'level_styles': ls,
#        'longest_level': longest_level,
#        'key_styles': ks,
#        'justed_levels': justed_levels,
#    }
#

# -------------------------------------------------------- oldish 8 color stuff
# TODO gk: check if we can ditch all below:
class ShellColors:
    K = '0;30'
    B = '0;34'
    G = '0;32'
    C = '0;36'
    R = '0;31'
    P = '0;35'
    O = '0;33'
    Y = '1;33'
    W = '1;37'
    # bold:
    BK = '1;30'
    BW = '0;37'
    BB = '1;34'
    BG = '1;32'
    BC = '1;36'
    BR = '1;31'
    BP = '1;35'
    N = '0'
    # distingishable cols, to count up numbers:
    cols = [
        'BG',
        'BB',
        'BW',
        'BC',
        'BR',
        'BP',
        'BK',
        'G',
        'P',
        'B',
        'C',
        'R',
        'P',
        'O',
        'Y',
        'W',
    ]


def colors():
    ret = OrderedDict()
    for c in ShellColors.cols:
        ret[c] = getattr(ShellColors, c)
    return ret


def esc_code(fg, bg=None, for_ps1=0):
    l = len(ShellColors.cols)
    if str(fg).isdigit():
        fg = ShellColors.cols[int(fg) % l]
    fg = '\\033[' + getattr(ShellColors, fg) + 'm'
    if bg is not None:
        if str(bg).isdigit():
            bg = ShellColors.cols[int(bg) % l]
        bg = getattr(ShellColors, bg)
        # no bold for bgs, so no 0; or 1;:
        bg = '\\033[4' + bg[-1] + 'm'
    esc = ''
    end_esc = ''
    if for_ps1:
        esc = '\\['
        end_esc = '\\]'
    ret = esc + fg
    if bg is not None:
        ret += esc + bg
    return ret + end_esc


def colorize(s, fg, bg=None, for_ps1=0):
    ret = '%s%s%s' % (esc_code(fg, bg, for_ps1), s, esc_code('N', bg, for_ps1))
    return ret


def colortable():
    for i in range(1, 256):
        print('%6s' % i, '\x1b[38;5;%smthis is color number %s\x1b[0m' % (i, i))


if __name__ == '__main__':
    colortable()
    print()
    Theme.print()
    print(I('With Environment Ignored'))
    Theme.setup(ignore_env=False)
    Theme.print()



"""
from build/bin/<app run script>
- for out of process entry
- early load of app module, for gevent and other imports

"""
import importlib
import json
import os
import sys
import traceback

py_env = {}
app_mod = None
app_type = None  # 'subproc|module'
app_import_err = None
enter_path = os.environ['PATH']


def app_load(fn_env):
    if py_env:
        return
    with open(fn_env) as fd:
        e = json.loads(fd.read())
    py_env.update(e)
    for p in reversed(e.get('PYTHONPATH', '').split(':')):
        sys.path.insert(0, p)
    [sys.path.append(p) for p in e.get('sys_path_append', '').split(':')]
    # all subprocs need this:
    export_env()
    if_app_is_module_then_load_it()


def export_env():
    m = {}
    for k, v in py_env.items():
        if isinstance(v, str):
            m[k] = v
            continue
        m[k] = json.dumps(v)
    os.environ.update(m)


def if_app_is_module_then_load_it():
    """ we remember the finding in a var file app_type"""
    global app_type
    app = py_env.get('app')
    if not app:
        raise Exception('No $app')

    def try_load():
        global app_mod, app_type, app_import_err
        if '/' in app:
            app_type = 'subproc'
            return
        app_type = 'module'
        # probably... not sure.
        # but we see now - early import in any case (gevent..)
        try:
            app_mod = importlib.import_module(app)
        except ImportError as ex:
            r = traceback.extract_tb(sys.exc_info()[2])[4:]
            l = [l.filename for l in r if '/%s' % app in l.filename]
            if not l:
                # this not a module, which just has an import error
                app_type = 'subproc'
                return
            # will report when we have logging:
            app_import_err = ex
        except Exception as ex:
            print('wtf', ex)
            breakpoint()
            i = 1

    # Trying to remember the app type is too risky: One wrong config and it never works again:
    try_load()
    return

    # ft = py_env['var_dir'] + '/app_type'
    # if os.path.exists(ft):
    #    with open(ft) as fd:
    #        app_type = fd.read().strip()
    # if app_type == None:
    #    try_load()
    #    with open(ft, 'w') as fd:
    #        fd.write(app_type)
    # elif app_type == 'module':
    #    try_load()



import os
import uuid
from fnmatch import fnmatch
from functools import partial

import jwt
from absl import flags

FLAGS = flags.FLAGS

flags.DEFINE_string('token_name', 'X-AX-API-TOKEN', 'Token name')
flags.DEFINE_boolean('no_token_check', False, 'Omit token checks (for dev mode)')
flags.DEFINE_string(
    'token_secret',
    '%(DA_DIR_SECURE)s/%(name)s/.token_secret',
    'filename, $.. or direct',
)


def read_token(realm, raw_token=None, get_app=None, secret=None, unauth=None):
    app = get_app()
    app_server = app['app_server']
    raw_token = (
        app_server.request.headers.get(FLAGS.token_name, 'x')
        if raw_token is None
        else raw_token
    )
    app_server.response.headers['Cache-Control'] = 'no-store'
    app_server.response.headers['Server'] = '%(name)s/%(version)s' % app
    if FLAGS.no_token_check:
        token = {'realms': '*', 'roles': 'publ_*'}
        raw_token = str(uuid.uuid4())
    else:
        try:
            token = jwt.decode(raw_token, secret, algorithms=['HS256'])
            realms = token['realms']
            # token = {'token': t}
        except:
            unauth('Require valid %s' % FLAGS.token_name)

    realms = token['realms']
    if not fnmatch(realm, realms + '*'):
        return unauth('Token invalid (realm)')

    return token, raw_token


# ------------------------------------------------------------------- app setup
def read_secret(get_app, unauth):
    if FLAGS.no_token_check:
        secret = ''
    else:
        s = FLAGS.token_secret
        m = dict(os.environ)
        m['name'] = get_app()['name']
        if '%(' in s:
            s = s % m
        if os.path.exists(s):
            s = 'file://' + s
        if not s.startswith('file://'):
            if s.startswith('$'):
                secret = os.environ[s[1:]]
            else:
                secret = s
        else:
            # for now:
            fn = s.split('file://', 1)[1]
            if not os.path.exists(fn):
                log.warn('token secret file not found', filename=fn)
                sys.exit(1)
            with open(fn) as fd:
                secret = fd.read().strip()
    return partial(read_token, get_app=get_app, unauth=unauth, secret=secret)



# coding:utf-8
import re

from .inflect import (
    camelize,
    dasherize,
    humanize,
    ordinal,
    ordinalize,
    parameterize,
    pluralize,
    singularize,
    titleize,
    transliterate,
    underscore,
)


def convert_keys(m, to=None, recurse=1, in_place=0):
    """ The main function to be used.
    Converts the structure m to intended key format
    The conversion is NOT done inplace, new structures are returned

    in_place: We replace the keys of the given structure

    to: None -> we return available converters
        (custom_callable, 'camel') -> call the custom one, then to camel
        'snake, title' -> convert to snake, then titlelize


    in_place: 0 -> new structure returned, type type of old
              1 -> replacement of keys
              2 -> adding of converted keys
    """
    # no conversion of simple types, these are normally values and not structs
    # with keys:
    if isinstance(m, (basestring, bool, int, float)):
        return m
    if not to:
        return CONVERTERS.keys()
    if isinstance(to, basestring):
        to = to.split(',')
    for _to in to:
        if isinstance(_to, basestring):
            _to = _to.strip()
            if not _to:
                continue
            converter = CONVERTERS[_to]
        else:
            # converter is given as callable:
            converter = _to
        m = _convert_keys(m, converter, recurse, in_place)
    return m


def _convert_keys(m, converter, recurse, in_place):
    if isinstance(m, tuple):
        # in place makes no sense here but we rather don't except but return a
        # new one
        l = ()
        for k in m:
            if isinstance(k, dict):
                l += (
                    _convert_keys(
                        k, recurse=recurse, converter=converter, in_place=in_place
                    ),
                )
            else:
                l += (k,)
        return l

    if isinstance(m, list):
        if not in_place:
            l = type(m)()
            l.extend(m)
        else:
            l = m
        for i in xrange(len(l)):
            k = l[i]
            if isinstance(k, dict):
                l[i] = _convert_keys(
                    k, recurse=recurse, converter=converter, in_place=in_place
                )
        return l

    if isinstance(m, dict):
        if not in_place:
            n = type(m)()
        else:
            n = m
        for k, v in m.items():
            if recurse and isinstance(v, (list, tuple, dict)):
                v = _convert_keys(v, recurse=1, converter=converter, in_place=in_place)
            if in_place == 1:
                m.pop(k)
            k = converter(k)
            n[k] = v
        return n
    else:
        # string, int?
        return converter(m)


def camel_1(string):
    """ this does numbers seperated, Foo12Bar -> foo_12_bar """
    string = string.replace(' ', '')
    string = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', string)
    string = re.sub('(.)([0-9]+)', r'\1_\2', string)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', string).lower()


def camel(string, uc=False):
    if not isinstance(string, basestring):
        return string
    string = string.replace(' ', '_')
    # underscore first to prevent 'HTTPAuthPassword -> hTTPAuthPassword:
    return camelize(underscore(string), uppercase_first_letter=uc)


def upper_camel(string):
    return camel(string, uc=True)


def ordinal(nr):
    if not isinstance(nr, int):
        return nr
    return ordinalize(nr)


def snake(string):
    if not isinstance(string, basestring):
        return string
    string = string.replace(' ', '_')
    return underscore(string)


def title(string):
    if not isinstance(string, basestring):
        return string
    return titleize(string)


def plural(string):
    if not isinstance(string, basestring):
        return string
    return pluralize(string)


def ascii(string):
    if not isinstance(string, basestring):
        return string
    return str(transliterate(unicode(string)))


def snake_plural(string):
    return plural(snake(string))


def singular(string):
    if not isinstance(string, basestring):
        return string
    return singularize(string)


def param_dash(string, d='-'):
    if not isinstance(string, basestring):
        return string
    return str(parameterize(unicode(string), d))


def param_under(string):
    return param_dash(string, d='_')


def param_space(string):
    return param_dash(string, d=' ')


def human(string):
    if not isinstance(string, basestring):
        return string
    # want fooBar -> Foo bar
    return humanize(snake(string))


def dash(string):
    if not isinstance(string, basestring):
        return string
    return dasherize(snake(string))


# ----------------------------------------- All funcs below compat with old API:
def camel_to_snake(name):
    return camel_1(name)


under_pat = re.compile(r'_([a-z])')


def snake_to_camel(name):
    return under_pat.sub(lambda x: x.group(1).upper(), name)


def convert(s, remove_spaces=1):
    a = re.compile('((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))')
    res = a.sub(r'_\1', s).lower()
    # keys like 'Foo Bar' -> foo_bar
    if remove_spaces:
        res = res.replace(' ', '')
    return res


def convertSnake(j):
    """ this is convert FROM snake to json(camel) """
    out = {}
    for k in j:
        new_k = k.split('_')
        out[new_k[0] + ''.join(x.title() for x in new_k[1:])] = j[k]
    return out


def convertJSON(j, remove_spaces=1):
    """ this is convert FROM JSON to snake """
    out = {}
    for k in j:
        newK = convert(k, remove_spaces)
        if isinstance(j[k], dict):
            out[newK] = convertJSON(j[k], remove_spaces)
        elif isinstance(j[k], list):
            out[newK] = convertArray(j[k], remove_spaces)
        else:
            out[newK] = j[k]
    return out


def convertArray(a, remove_spaces=1):
    newArr = []
    for i in a:
        if isinstance(i, list):
            newArr.append(convertArray(i))
        elif isinstance(i, dict):
            newArr.append(convertJSON(i))
        else:
            newArr.append(i)
    return newArr


l = list(globals())
CONVERTERS = {}
for k in l:
    if (
        k.startswith('convert')
        or k.startswith('_')
        or '_to_' in k
        or k.startswith('re')
        or k.endswith('ize')
    ):
        continue
    CONVERTERS[k] = globals()[k]

if __name__ == '__main__':
    print(convert_keys(''))

    def custom_conv(s):
        return 'Aaa'

    m = {'fooBar': 1}
    n = convert_keys(m, (custom_conv, 'snake'))
    assert n == {'aaa': 1}

    n = convert_keys(m, 'camel')
    # not in place, so:
    m[1] = 2
    assert m != n

    m = {'FooBar': 1}
    n = convert_keys(m, 'camel', in_place=1)
    assert len(n) == 1
    m[1] = 2
    assert n == m

    m = {'FooBar': 1}
    n = convert_keys(m, 'camel', in_place=2)
    assert len(n) == 2
    m[1] = 2
    assert n == m

    l = convert_keys([1, 2, m], 'snake')
    assert l == [1, 2, {1: 2, 'foo_bar': 1}]
    assert l[2] != m

    l1 = convert_keys([1, 2, m], 'snake', in_place=1)
    assert l1[2] == m
    assert l1 == l

    m = {'foo_bar': {'FooBar': 23, 23: 23}}

    n = convert_keys(m, 'camel')
    assert n == {'fooBar': {23: 23, 'fooBar': 23}}

    n = convert_keys(m, 'camel', recurse=0)
    assert n == {'fooBar': {23: 23, 'FooBar': 23}}

    n = convert_keys(m, 'ordinal')
    assert n == {'foo_bar': {'23rd': 23, 'FooBar': 23}}

    n = convert_keys(m, 'camel, ordinal')
    assert n == {'fooBar': {'23rd': 23, 'fooBar': 23}}

    n = convert_keys(m, 'upper_camel')
    assert n == {'FooBar': {23: 23, 'FooBar': 23}}

    n = convert_keys(m, 'title')
    assert n == {'Foo Bar': {23: 23, 'Foo Bar': 23}}

    n = convert_keys(m, 'plural')
    assert n == {'foo_bars': {23: 23, 'FooBars': 23}}

    n = convert_keys(m, 'plural')
    assert n == {'foo_bars': {23: 23, 'FooBars': 23}}

    n = convert_keys(m, 'snake_plural')
    assert n == {'foo_bars': {'foo_bars': 23, 23: 23}}
    assert plural('category') == 'categories'

    pl = convert_keys(m, 'plural')
    n = convert_keys(pl, 'singular')
    assert n == m

    u = {'Donald E. Knuth': {'!foo': 23}}
    assert convert_keys(u, 'param_dash') == {'donald-e-knuth': {'foo': 23}}

    u = {'Donald E. Knuth': {'!foo': 23}}
    assert convert_keys(u, 'param_under') == {'donald_e_knuth': {'foo': 23}}

    u = {'Donald E. Knuth': {'!foo': 23}}
    assert convert_keys(u, 'param_space') == {'donald e knuth': {'foo': 23}}

    n = convert_keys(m, 'human')
    assert n == {'Foo bar': {23: 23, 'Foo bar': 23}}

    n = convert_keys(m, 'dash')
    assert n == {'foo-bar': {23: 23, 'foo-bar': 23}}

    u = {'älämölö': {'Ærøskøbing': 23, 2: 23}}
    assert convert_keys(u, 'ascii') == {'alamolo': {2: 23, 'rskbing': 23}}

    test_cases = list()
    test_cases.append(('camelCase', 'camel_case'))
    test_cases.append(('camelCaseCase', 'camel_case_case'))
    test_cases.append(('camel2Case', 'camel_2_case'))
    test_cases.append(('camel12Case', 'camel_12_case'))
    test_cases.append(('camel12Case', 'camel_12_case'))
    test_cases.append(('camelCaseURL', 'camel_case_url'))
    test_cases.append(('camel2CaseURL', 'camel_2_case_url'))
    test_cases.append(('camel12CaseURL', 'camel_12_case_url'))
    test_cases.append(('camel12Case2URL', 'camel_12_case_2_url'))
    test_cases.append(('camel12Case12URL', 'camel_12_case_12_url'))
    test_cases.append(('CamelCase', 'camel_case'))
    test_cases.append(('CamelCaseCase', 'camel_case_case'))
    test_cases.append(('URL CamelCase', 'url_camel_case'))
    test_cases.append(('Spaced URL Casea', 'spaced_url_casea'))
    for string, res in test_cases:
        print(string + ' -> ' + camel_1(string))
        try:
            assert camel_1(string) == res
        except Exception as ex:
            import pdb

            pdb.set_trace()

    assert camel_to_snake('asdf') == 'asdf'
    assert camel_to_snake('asdfAsBs') == 'asdf_as_bs'
    assert camel_to_snake('AsdfAsBs') == 'asdf_as_bs'
    assert camel_to_snake('fooBar') == 'foo_bar'
    assert snake_to_camel('foo_bar') == 'fooBar'

    m = {
        'someObject': [
            {'anotherObject': 'CamelCaseValue'},
            {'anotherObject': 'AnotherCamelCaseValue'},
        ]
    }
    assert convertJSON(m) == {
        'some_object': [
            {'another_object': 'CamelCaseValue'},
            {'another_object': 'AnotherCamelCaseValue'},
        ]
    }



# -*- coding: utf-8 -*-
import io
import sys

PY2, U8 = sys.version_info[0] < 3, 'UTF-8'
if PY2:

    def unic(s):
        return s if isinstance(s, unicode) else unicode(s, encoding=U8)

    def bytes_(s):
        return s if isinstance(s, str) else s.encode(U8)


else:

    def unic(s):
        return s if isinstance(s, str) else str(s, encoding=U8)

    def bytes_(s):
        return s if isinstance(s, bytes) else bytes(s, encoding=U8)


native = bytes_ if PY2 else unic


# http://jrgraphix.net/r/Unicode/2500-257F
# e.g.: [print(chr(i)) for i in box_drawing]
class blocks:
    """usage: e.g. term.py"""

    geometric_shapes = range(0x25A0, 0x25FF)
    box_drawing = range(0x2500, 0x257F)
    arrows = range(0x2190, 0x21FF)
    basic_latin = range(0x0020, 0x007F)


def read_uni_chars_off_str(s, chars, buffer=None):
    fd = io.BytesIO(bytes_(s))
    res = read_uni_chars(fd, chars)
    fd.close()
    return res


def read_uni_chars(fd, chars, buffer=None):
    """
    read as many bytes from an fd as required to get <chars> characters back -
    as str in 2, as unicode in3
    """
    if not PY2:
        # fd must be a textio wrapper, why would you open an fd with 'rb' in
        # py3 if you want the chars?
        return fd.read(chars)

    if not chars > 0:
        return fd.read(chars)

    # now the problem: Py2, user wants 10 chars:
    # could not get this to work with BytesIO/TextIO because its late but
    # there Must be a way -> remove this crap when fit:
    buffer = buffer or chars
    s = unicode('')

    s, last_chunk_bytes = unicode(''), ''

    while True:
        chunk = fd.read(buffer)
        if not chunk:
            break
        chunk, last_chunk_bytes = split_off_incompl(chunk, last_chunk_bytes)
        if not chunk and not last_chunk_bytes:
            break
        s += chunk
        if len(s) >= chars:
            break
    return s.encode('utf-8') if PY2 else s


def split_off_incompl(chunk, pre=None):
    """
    returns the decodable part of a chunk, plus the last bytes which
    are incomplete
    """

    def is_compl(b):
        try:
            return b.decode('utf-8')
        except:
            return False

    if pre is not None:
        chunk = pre + chunk

    last = bytes_('')
    while True:
        uni = is_compl(chunk)
        if uni is not False:
            return uni, last
        chunk, l = chunk[:-1], chunk[-1]
        last = l + last


# ---------------------------------------------------------- oldish stuff below
# remove when surely uneeded
class UCs:
    """ unicode characteres. type new ones in vi (insert mode) like:
        strg-v u 2500
    """

    # for table prettyfying:
    VU = '│'
    H = '─'
    JUNC = '┼'

    NOTE = '♩'
    NOTES = '♬'
    PIEK = '♠'
    MARK = '☀'
    SKULL = '☠'
    GOOD = '☺'
    BAD = '☹'
    # not curses complient, i. not in bpython:
    EMO_SMILE = '😀'


if __name__ == '__main__':
    print('─')
    for k in dir(UCs):
        print(getattr(UCs, k))



"""
SuperLight Structured Logging
-----------------------------

Stdlib Free Fast Logging, Using Structlog Only

Usage:
    axlog.setup_logging()  once per app
    axlog.get_logger('axwifi').info('foo', bar=baz)

    The app may use flags: --log_level=20
"""


import os
import sys
import traceback

import structlog
from devapp.tools import FLG, define_flags
from pygments.styles import get_all_styles
from structlog import BoundLoggerBase, PrintLogger, wrap_logger
from structlog.exceptions import DropEvent

# -- Setup for a stdlib logging free, getattr free use:
from structlog.processors import JSONRenderer
from structlogging import processors as ax_log_processors
from structlogging import renderers

pygm_styles = list(get_all_styles())
pygm_styles.extend(['light', 'dark'])


class flags:
    autoshort = ''

    class log_level:
        n = 'Log level (10: debug, 20: info, ...). You may also say log_level=error'
        d = '20'

    class log_time_fmt:
        n = 'Log time format. Shortcuts: "ISO", "dt"'
        d = '%m-%d %H:%M:%S'

    class log_fmt:
        n = 'Force a log format. 0: off, 1: auto, 2: plain, 3: plain_no_colors, 4: json. '
        n += 'Note: This value can be set away from auto via export log_fmt as well.'
        d = 'auto'

    class log_add_thread_name:
        n = 'Add name of thread'
        d = False

    class log_dev_fmt_coljson:
        n = 'List of keys to log as json.'
        d = 'json,payload'
        t = list

    class log_dev_coljson_style:
        n = 'Pygments style for colorized json. To use the 16 base colors and leave it '
        n += 'to the terminal palette how to render choose light or dark'
        t = sorted(pygm_styles)
        d = 'dark'


define_flags(flags)

# fmt:off
log_levels = {
    'fatal'     : 70,
    'critical'  : 60,
    'error'     : 50,
    'exception' : 50,
    'err'       : 50,
    'warn'      : 40,
    'warning'   : 40,
    'info'      : 20,
    'debug'     : 10,
}
# fmt:on
def storing_testlogger(dub_to=None, store=None):
    """for tests - stores into array, optionally dubs to existing logger"""
    store = [] if store is None else store

    def add_log_to_store(logger, _meth_, ev, store=store, log=dub_to):
        d = dict(ev)
        d['_meth_'] = _meth_
        store.append(d)
        if log:
            getattr(log, _meth_)(ev.pop('event'), **ev)
        raise structlog.DropEvent

    gl = structlog.get_logger
    l, m = ('dub', gl) if dub_to is None else (dub_to, structlog.wrap_logger)
    r = m(l, processors=[add_log_to_store])
    r.reset = r.clear = lambda s=store: s.clear()
    r.store = store
    return r


def log_flags():
    flog = get_logger('flag')
    m = FLG.flag_values_dict()
    for k in sorted(m.keys()):
        flog.debug('', **{k: m[k]})


class AXLogger(BoundLoggerBase):
    """A Bound Logger is a concrete one, e.g. stdlib logger"""

    def log(self, event, kw, _meth_od):
        # the frame will allow *pretty* powerful parametrization options
        # todo: allow that for the user
        # the lambda below is also a frame:
        kw['_frame_'] = sys._getframe().f_back.f_back
        try:
            args, kw = self._process_event(_meth_od, event, kw)
        except DropEvent:
            return
        return self._logger.msg(*args, **kw)


class AppLogLevel:
    """Intermediate resets of devapp app loglevel

    with AppLogLevel(30):
        do_lots_of_stuff()

    """

    def __init__(self, level):
        from devapp.app import app

        self.app = app
        self.level = level
        self.was_level = app.log._logger.level

    def __enter__(self):
        self.app.log._logger.level = self.level

    def __exit__(self, *a, **kw):
        self.app.log._logger.level = self.was_level


# setting all level methods, like log.warn into the logger:
for l, nr in log_levels.items():
    setattr(AXLogger, l, lambda self, ev, _meth_=l, **kw: self.log(ev, kw, _meth_))


def get_logger(name, level=None, **ctx):
    """supports name and level conventions - matches the processor chain"""
    if not structlog.is_configured():
        setup_logging()

    FLG.log_level = int(log_levels.get(FLG.log_level, FLG.log_level))
    level = level or FLG.log_level
    log = wrap_logger(PrintLogger(file=sys.stderr), wrapper_class=AXLogger)
    log._logger.name = name
    log._logger.level = level
    return log.bind(**ctx)


def filter_by_level(logger, _meth_, ev):
    if log_levels[_meth_] < logger.level:
        raise DropEvent
    return ev


censored = ['password', 'access_token']


def censor_passwords(_, __, ev, _censored=censored):
    """
    Supports to deliver the censor structure in ev or at setup:
    log.info('MyMsg', data={'foo': {'pw': ..}}, censor=('token', ('data', 'pw')))
    """
    censored = ev.pop('censor', _censored)
    for pkw in censored:
        # nested censored value?
        if isinstance(pkw, (list, tuple)):
            pw = ev
            for part in pkw:
                try:
                    pw = pw.get(part)
                except Exception as ex:
                    pw = None
                if not pw:
                    break
        else:
            pw = ev.get(pkw)
        if pw:
            # TODO: does only work with ONE censored field
            k = min(len(pw) - 1, 3)
            v = pw[:k] + '*' * min((len(pw) - k), 10)
            if isinstance(pkw, (list, tuple)):
                cur = ev
                for part in pkw[:-1]:
                    cur = ev[part]
                cur[pkw[-1]] = v

            else:
                ev[pkw] = v
    return ev


def frame(f):
    m = {}
    m['pos'] = [f.lineno, f.filename]
    m['line'] = f.line
    # if not f.name == '<lambda>':
    m['name'] = f.name
    return m


def add_exception_stack(_, __, ev):
    """This filters out huge rx stacks, which are not useful at all normally.
    """
    if 'exc' in ev:
        E = ev['exc']
        if isinstance(E, Exception):
            r = traceback.extract_tb(sys.exc_info()[2])
            ev['stack'] = [
                'Traceback',
                [frame(l) for l in r if not '/rx/' in l.filename],
                [E.__class__.__name__, E.args],
            ]
    return ev


def add_stack_info(_, __, ev):
    # from structlog._frames import _find_first_app_frame_and_name as ffafan

    si = ev.pop('stack_info', 0)
    if si and not 'stack' in ev:
        # TODO make the non raise stack_info tbs also better, like the exc ones
        ev['stack'] = traceback.format_stack(ev['_frame_'][0])
    return ev


def add_logger_name_and_frame_infos(logger, _, ev):
    # TODO parametrizeable frame info extraction.
    ev.pop('_frame_')
    ev['logger'] = '%s' % (logger.name)
    return ev


std_processors = [
    filter_by_level,
    # structlog.stdlib.add_logger_name,
    structlog.stdlib.add_log_level,
    censor_passwords,
    add_exception_stack,
    add_stack_info,  # allows to say stack_info=1 w/o exceptions
    add_logger_name_and_frame_infos,
]


def fmt_to_int(fmt):
    if str(fmt) in ('auto', '1'):
        fmt = os.environ.get('log_fmt', fmt)
    if fmt.isdigit():
        return int(fmt)
    m = {'off': 0, 'auto': 1, 'plain': 2, 'plain_no_colors': 3, 'json': 4}
    return m.get(fmt, 1)


def log_dropper(*a):
    raise structlog.DropEvent


# class F:
#     log_fmt = 'plain'
#     log_level = 10
#     log_time_fmt = '%m-%d %H:%M:%S'
#     log_add_thread_name = True
#     log_dev_fmt_coljson = []
#     log_dev_coljson_style = ('paraiso-dark',)
import ujson, json, time


def to_str(obj):
    try:
        return obj.decode('utf-8')
    except:
        return str(obj)


def safe_dumps(obj, to_str=to_str, default=None):
    try:
        # fails when objects are in, e.g. connection sockets:
        # can dump tuple keys
        return ujson.dumps(obj, ensure_ascii=False, reject_bytes=False)
    except:
        try:
            return json.dumps(obj, default=default)
        except Exception as ex:
            return json.dumps(
                {
                    'event': 'cannot log-serialize: %s' % str(obj),
                    'level': 'error',
                    'logger': 'sl',
                    'timestamp': time.time(),
                }
            )


def setup_logging(
    name=None,
    level=None,
    processors=None,
    censor=(),
    # these come from the app. should not overwrite flags except if not given
    log_dev_fmt_coljson=None,
    log_time_fmt=None,
    get_renderer=False,
):
    """
    pygmentize: Those keys, if present, will be rendered as colored json
    This requires app.run done.
    If not: Say FLG(sys.argv)
    """
    # try:
    #     log_fmt = FLG.log_fmt
    # except:
    #     breakpoint()  # FIXME BREAKPOINT
    #     FLG = F

    # are we a real app, or just a test program:
    # fmt:off
    log_fmt               = FLG.log_fmt
    log_time_fmt          = FLG.log_time_fmt
    log_add_thread_name   = FLG.log_add_thread_name
    log_dev_coljson_style = FLG.log_dev_coljson_style
    log_dev_fmt_coljson_flg  = FLG.log_dev_fmt_coljson
    # fmt:on

    if censor:
        censor = [censor] if isinstance(censor, str) else censor
        [censored.append(c) for c in censor if not c in censored]
    fmt_vals, val_formatters = {}, {}
    # log.info('Response', json=... ) auto formats:
    lc = log_dev_fmt_coljson_flg
    if log_dev_fmt_coljson:
        lc.extend(log_dev_fmt_coljson)
    for k in lc:
        fmt_vals[k] = 'f_coljson'
    fmt_vals['stack'] = 'f_coljson'  # stack tracebacks always

    val_formatters['f_coljson'] = {
        'formatter': 'coljson',
        'style': log_dev_coljson_style,
    }

    fmt = fmt_to_int(log_fmt)
    tty = sys.stdout.isatty()
    if fmt == 4 or (fmt == 1 and not tty):
        rend = JSONRenderer(serializer=safe_dumps)
    else:
        # console renderer. colors?
        colors = fmt != 3
        rend = renderers.ThemeableConsoleRenderer(
            colors=colors, fmt_vals=fmt_vals, val_formatters=val_formatters
        )

    if get_renderer:
        return rend

    p = processors or std_processors
    if log_add_thread_name:
        p.insert(1, ax_log_processors.add_thread_name)
    if log_time_fmt:
        p.insert(1, ax_log_processors.TimeStamper(fmt=log_time_fmt))
    p.append(rend)
    if fmt == 0:
        p.insert(0, log_dropper)

    structlog.configure(
        processors=p,
        context_class=dict,
        wrapper_class=AXLogger,
        cache_logger_on_first_use=True,
    )
    # all in one:
    if name is not None:
        return get_logger(name, level=level)


# .



# TODO DEPRECATED - move to archive
"""
CLI Wrapper.
- Maps CLI to Class Trees (namespaces) with functions

- Is able to pull out the devapp switches before digging into the tree

"""

import sys
from functools import partial
from inspect import signature

import mdvl
from devapp.app import app, run_app


def auto_type(v):
    for k, b in ('true', True), ('false', False):
        if str(v).lower() == k:
            return b
    for t in float, int:
        try:
            return typ(v)
        except:
            pass
    if not t:
        return ''
    return str(v)


def not_found(*a, **kw):
    app.die('Not supported')


def find_in_cls(cls, p):
    v = getattr(cls, p, None)
    if v is not None:
        return v
    p = p.lower()
    for k in dir(cls):
        if k.lower().startswith(p):
            vv = getattr(cls, k)
            if callable(vv):
                return vv


def sh_help(cls, cli, out=None, hir=0):
    hir += 1
    out = [] if out is None else out
    out.append('%s %s' % ('#' * hir, cls.__name__))
    d = getattr(cls, '__doc__')
    if d and '-hh' in cli:
        out.append(d)
    dflt_meth = getattr(cls, 'default', None)
    clss = []
    for k in dir(cls):
        if k.startswith('_') or k == 'default':
            continue
        c = getattr(cls, k)

        if isinstance(c, type):
            clss.append(c)

        elif callable(c):
            if c == dflt_meth:
                k += '[default]'
            out.append(
                '- **%s**(%s  '
                % (k, str(signature(c)).split('(', 1)[1].replace('**', '..'))
            )
            d = getattr(c, '__doc__')
            if d and '-hh' in cli:
                out.append(d)
    for c in clss:
        sh_help(c, cli, out, hir)
    if hir == 1:
        mdvl.render('\n'.join(out))
        return 0


def run_cls_tree(cls_tree):
    """sys.argv = [app, devappflags.., cls, subcls..., func, funcargs...]"""
    # get the devapp run flags into sys.argv (unit first non minus arg):
    for i in range(1, len(sys.argv)):
        if not sys.argv[i].startswith('-'):
            break
    sys.argv, meth_args = sys.argv[:i], sys.argv[i:]

    # get command func:
    cmd = cls_tree
    while meth_args:
        # get the command function from the nested class tree:
        p = meth_args.pop(0)
        ncmd = find_in_cls(cmd, p)
        if ncmd == None:
            meth_args.insert(0, p)
            break
        cmd = ncmd

    # we do our own help system, this is not about flags:
    b = [*sys.argv, *meth_args]
    if '-h' in b or '--help' in b or '-hh' in b:
        sys.exit(sh_help(cmd, cli=b))

    if isinstance(cmd, type):
        cmd = getattr(cmd, 'default', not_found)

    # get command args:
    a, kw = (), {}
    for p in meth_args:
        if '=' in p:
            k, v = p.split('=', 1)
            kw[k] = auto_type(v)
        else:
            a += (p,)
    # wrap command function with args:
    cmd = partial(cmd, *a, **kw)
    # kw_log = dict(
    #    log_dev_fmt_coljson=['res'],
    #    censor=(['res', 'token'], ['res', 'access_token']),
    # )
    return run_app(cmd)  # , kw_log=kw_log)



#!/usr/bin/env python
"""
Since entr always requires 2 terminals, one for entr, one for the reloader,
we've put filewatching capabilities into the devapp itself - spawned as back
ground process - this one.

Usage:
    See devapp.app.py, search dirwatch

"""

import os
import sys
import time
from fnmatch import fnmatch
from functools import partial

WD = 'WATCHDOG: '

out = partial(print, file=sys.stderr)


def start_dir_watch(dir_pid_match_rec):
    # print(dir_pid_match_rec)
    dir, pid, match, recursive = dir_pid_match_rec.split(':')
    recursive = bool(recursive)
    pid = int(pid)
    if not '*' in match:
        match = ('*' + match + '*').replace('**', '*')
    l = dict(locals())
    l.pop('dir_pid_match_rec')

    out(WD + 'starting. %s' % str(l)[1:-1].replace("'", ''))
    from watchdog.observers import Observer
    from watchdog.events import FileSystemEventHandler

    class H(FileSystemEventHandler):
        def on_modified(self, event, pid=pid):
            out(WD + f'event: {event.event_type}  path : {event.src_path}')
            # out(WD + 'match' + self.match + '.')
            if self.match:
                if not fnmatch(event.src_path, self.match):
                    return
            out(WD + 'match => Sending reload')
            os.kill(int(pid), 1)

    o, h = Observer(), H()
    if isinstance(match, str):
        h.match = match
    else:
        raise Exception('spec not supported')

    o.schedule(h, path=dir, recursive=recursive)
    o.start()
    while 1:
        time.sleep(100000)


if __name__ == '__main__':
    start_dir_watch(sys.argv[1])



# coding: utf-8
import unittest

# from ax.utils.parsing.parse_text import *
from collections import OrderedDict

from theming import unicode_chars as unic
from theming.pretty_print import *

PY2 = sys.version_info[0] < 3
LS = '\n'


def unspace(s):
    return s.strip().replace(' ', '').replace('\t', '')


def clean(s):
    return unspace(s).replace('\n', '')


class TestUnicode(unittest.TestCase):
    def test_unisplit(self):

        # cannot make an incomplete bytearray witout using b which crashes
        # the python2 compiler. The function is anyway only really needed in2
        if not PY2:
            return 'skipped'
        s = '① ② ③ ④ ⑤ ⑥ ⑦ ⑧ ⑨ ⑩'
        b = unic.bytes_(s) + '\x01'
        res, b1 = unic.split_off_incompl(b)
        assert type(res) == type(unic.unic(''))
        assert b1 == ''

        b = unic.bytes_(s) + '\xe2\x91'
        res, b1 = unic.split_off_incompl(b)

        assert type(res) == type(unic.unic(''))
        assert unic.unic(res) == unic.unic(s)
        assert b1 == unic.bytes_('\xe2\x91')

    def test_str_split(self):

        # cannot make an incomplete bytearray witout using b which crashes
        # the python2 compiler. The function is anyway only really needed in2
        if not PY2:
            return 'skipped'
        s = '① ② ③ ④ ⑤ ⑥ ⑦ ⑧ ⑨ ⑩'

        b = unic.bytes_(s) + '\xe2\x91'
        with self.assertRaises(Exception):
            u = unic.unic(b)
        eq = self.assertEqual
        for i in 0, 1, 2, 10, 100:

            res = unic.read_uni_chars_off_str(b, i)

            orig = unic.unic(s)
            orig_pre = orig[:i]
            orig_native = unic.native(orig_pre)
            eq(res, orig_native)
            assert type(res) == type('')


class TextFormatting(unittest.TestCase):
    def setUp(self):
        self.eq = self.assertEqual
        self.deq = self.assertDictEqual

        address = OrderedDict()
        address['bytes'] = 1.1
        address['city'] = 'anycity'
        address['on'] = True
        address['postal'] = 1234
        address['street'] = 'some road'

        data = OrderedDict()
        data['address'] = address
        data['id'] = [1, 'someid']
        data['title'] = 'home'

        self.z = {'data': data}

    def test_simple(s):
        z = {}
        ax = dict_to_txt(z, fmt={'ax': 1})
        assert type(ax) == str and not ax

    def test_list(s):
        z = ['axtest']
        ax = dict_to_txt(z, fmt={'ax': 1})
        assert ax == 'axtest\n-\n'

    def test_simple2(s):
        z = 'axtest'
        ax = dict_to_txt(z, fmt={'ax': 1})
        assert ax == 'axtest\n'

    def test_tuple(s):
        t = [1, [2, 3, 4], 3]
        assert obj_to_html_list(t) == (
            '\n<ul>\n<li>1</li>\n<li>\n<ul>\n<li>2</li>\n<li>3</li>'
            '\n<li>4</li>\n</ul>\n</li>\n<li>3</li>\n</ul>\n'
        )

    def test_map1(s):
        ax = dict_to_txt(s.z, fmt={'ax': 1})
        assert unspace(ax) == unspace(
            '''
data:
    address:
        bytes:  1.1
        city:   anycity
        on: True
        postal: 1234
        street: some road
    id:
        1
        someid
    title:  home
'''
        )

    def test_html(s):
        """ s.z =
{'data': {'address': {'bytes': 1.1,
                      'city': 'anycity',
                      'on': True,
                      'postal': 1234,
                      'street': 'some road'},
          'id': [1, 'someid'],
          'title': 'home'}}
        """
        assert clean(obj_to_html_list(s.z)) == clean(
            '''
<ul>
    <li>data:</li>
        <ul>
            <li>address:</li>
            <ul>
                <li>bytes: 1.1</li>
                <li>city: anycity</li>
                <li>on: True</li>
                <li>postal: 1234</li>
                <li>street: some road</li>
            </ul>
            <li>id:</li>
            <ul>
                <li>1</li>
                <li>someid</li>
            </ul>
            <li>title: home</li>
        </ul>
</ul>
'''
        )

    def test_plain(s):
        assert unspace(dict_to_txt(s.z)) == unspace(
            '''
data:
  address:
    bytes:1.1
    city:anycity
    on:True
    postal:1234
    street:some road
  id:
    1
    someid
  title:home
'''
        )
        s.assertNotIn(dict_to_txt(s.z), '\t')

    def test_plain_ax(s):
        ax = dict_to_txt(s.z, fmt={'ax': 1})
        assert unspace(ax) == unspace(
            '''
data:
    address:
        bytes:  1.1
        city:   anycity
        on: True
        postal: 1234
        street: some road
    id:
        1
        someid
    title:  home
'''
        )
        assert len(ax.split('\t')) == 27

    def test_tuple1(s):
        t = [1, [2, 3, 4], 3]
        assert (
            dict_to_txt(t, fmt={'ax': 1}).strip()
            == '''
1
-
2
3
4
-
3
'''.strip()
        )

    def test_wiki(s):
        p = dict_to_txt(s.z, fmt={'wiki': 1})
        assert (
            p.strip()
            == '''
 || data ||  ||
 ||  ||  || address ||  ||
 ||  ||  ||  || bytes || 1.1 ||
 ||  ||  || city || anycity ||
 ||  ||  || on || True ||
 ||  ||  || postal || 1234 ||
 ||  ||  || street || some road ||
 ||  || id ||  ||
 ||  ||  || 1 ||
 ||  ||  || someid ||
 ||  || title || home ||
'''.strip()
        )

    def test_html2(s):
        html = dict_to_txt(s.z, fmt={'html': 1})
        assert clean(html) == clean(
            '''
<table >
  <tr ><td >data</td><td ></td><td colspan="100" ></td></tr>
  <tr ><td ></td><td>
      <table >
        <tr ><td >address</td><td ></td><td colspan="100" ></td></tr>
        <tr ><td ></td><td>
            <table >
              <tr ><td >bytes</td><td >1.1</td><td colspan="100" ></td></tr>
              <tr ><td >city</td><td >anycity</td><td colspan="100" ></td></tr>
              <tr ><td >on</td><td >True</td><td colspan="100" ></td></tr>
              <tr ><td >postal</td><td >1234</td><td colspan="100" ></td></tr>
              <tr ><td >street</td><td >some road</td><td colspan="100" ></td></tr>
            </table>
            </td><td colspan="100" ></td></tr>
        <tr ><td >id</td><td ></td><td colspan="100" ></td></tr>
</td><td>
            <table >
              <tr ><td >1</td><td colspan="100" ></td></tr>
              <tr ><td >someid</td><td colspan="100" ></td></tr>
            </table>
            </td><td colspan="100" ></td></tr>
        <tr ><td >title</td><td >home</td><td colspan="100" ></td></tr>
      </table>
      </td><td colspan="100" ></td></tr>
</table>
'''
        )

    def test_html2_style(s):
        html = dict_to_txt(s.z, fmt={'html': 1, 'styles': {}})
        assert clean(html).startswith(
            clean(
                '''
<table class="table0">
  <tr class="tr0"><td class="td10">data</td><td class="td20"></td><td colspan="100" class="td30"></td></tr>
  <tr class="tr0"><td class="td10"></td><td>
      <table class="table1">
        <tr class="tr1"><td class="td11">address</td><td class="td21"></td><td colspan="100" class="td31"></td></tr>
        <tr class="tr1"><td class="td11"></td><td>
            <table class="table2">
'''
            )
        )

    def test_html_style(s):
        m = {'1': '2', 'sub': {1: '2', 2: 3}}
        html = dict_to_txt(m, fmt={'html': 1, 'styles': {'td': 'report3'}})
        assert clean(html) == clean(
            '''
<table class="table0">
  <tr class="tr0"><td class="report3">1</td><td class="report3">2</td><td colspan="100" class="report3"></td></tr>
  <tr class="tr0"><td class="report3">sub</td><td class="report3"></td><td colspan="100" class="report3"></td></tr>
  <tr class="tr0"><td class="report3"></td><td>
      <table class="table1">
        <tr class="tr1"><td class="report3">1</td><td class="report3">2</td><td colspan="100" class="report3"></td></tr>
        <tr class="tr1"><td class="report3">2</td><td class="report3">3</td><td colspan="100" class="report3"></td></tr>
      </table>
      </td><td colspan="100" class="report3"></td></tr>
</table>
'''
        )

    def test_mix(s):
        p2 = dict_to_txt([{2: [3, 3]}, {'sdf': 'asdfas'}], fmt={'ax': 1})
        assert p2 == '2:\t\n\t3\n\t3\n-\nsdf:\tasdfas\n'

    def test_tuple2(s):
        m = ['foo', ['foo', 'bar']]
        assert (
            dict_to_txt(m, fmt={'ax': 1}).strip()
            == '''
foo
-
foo
bar
    '''.strip()
        )

    def test_dict5(s):

        m = {
            'Operations': [
                {
                    'URL': 'http://your.url.com',
                    'Username': '',
                    'Password': '',
                    'UUID': '',
                    'ExecutionEnvRef': '',
                },
                {
                    'URL': 'http://your.url.com',
                    'Username': '',
                    'Password': '',
                    'Version': '',
                    'UUID': '',
                },
                {'Version': '', 'UUID': '', 'ExecutionEnvRef    ': ''},
            ],
            'CommandKey': '',
        }
        pretty = dict_to_txt(m, fmt={'ax': 1})
        assert (
            pretty
            == 'CommandKey:\t\nOperations:\t\n\tExecutionEnvRef:\t\n\tPassword:\t\n\tURL:\thttp://your.url.com\n\tUUID:\t\n\tUsername:\t\n\t-\n\tPassword:\t\n\tURL:\thttp://your.url.com\n\tUUID:\t\n\tUsername:\t\n\tVersion:\t\n\t-\n\tExecutionEnvRef    :\t\n\tUUID:\t\n\tVersion:\t\n'
        )


if __name__ == '__main__':
    unittest.main()



"""
Dependency free sourcer of shell style environ files into a
python process
"""


import json
import os
import subprocess as sp
import sys
from fnmatch import fnmatch
from os.path import exists

env = os.environ


def die(msg, **kw):
    print(msg, str(kw))
    sys.exit(1)


def source(*f, set_sys_path_from_py_path=None, ign_missing=False, insert_env_vars=()):
    """ sourcing shell config file
    we source all in one process, in order to allow referencing of vars within

    Referenced environ vars from outside are not resolved, unless state by insert_env_vars!
    """
    pyth = sys.executable

    F = []
    for i in list(f):
        if not exists(i):
            if not ign_missing:
                die('does not exist', fn=i)
        else:
            F.append(i)
    if not F:
        return {}
    # . is compatible with sh, source not
    src = '. ' + ' && . '.join(F)
    for i in insert_env_vars:
        if not '*' in i:
            src = ('export %s="%s" && ' % (i, env.get(i, ''))) + src
        else:
            ae = [k for k in env if fnmatch(k, i)]
            for j in ae:
                src = ('export %s="%s" && ' % (j, env.get(j, ''))) + src

    dump = '"%s" -Ssc "import os, json;' % pyth
    dump += 'print(json.dumps(dict(os.environ)))"'
    # -i: Starts with empty environ -> no refs work in our sourced file:
    cmd = ['env', '-i', '/bin/bash', '-c', 'set -a && %s && %s' % (src, dump)]
    out, err = sp.Popen(cmd, stdout=sp.PIPE, stderr=sp.PIPE).communicate()
    if err or not out:
        die('Error sourcing', f=f, err=err)
    res = json.loads(out)
    for k in 'SHLVL', '_', 'PWD':
        res.pop(k, 0)
    res = dict([(str(k), str(v)) for k, v in list(res.items())])
    if set_sys_path_from_py_path:
        pp = res.get('PYTHONPATH', '').split(':')
        for p in pp:
            sys.path.insert(0, p)
    return res



"""
Renders for a Type / Class Tree
"""

import os
from collections import OrderedDict as OD

from . import (
    TypeMetaClass,
    full_id,
    has_parent,
    props,
    simple_props,
    simple_types,
    walk_tree,
)

try:
    from colour import Color
except ImportError:
    Color = None
    # print('colour package missing - no hsl html rendering')

exists = os.path.exists


def color(hir):
    # hue, sat, lum
    return (
        hir * 50,
        min(2 * hir * 10 / 100.0, 0.8),
        min(2 * hir * 10 / 100.0, 0.8),
    )


def to_primitives(root, **cfg):
    t = '''
     new primitives.orgdiagram.ItemConfig({
                    id: %(id)s,
                    parent: %(parentid)s,
                    title: '%(type)s %(name)s %(level)s',
                    description: "%(hir)s %(name)s",
                    itemTitleColor: tinycolor("%(color)s").toHexString()
                }),
    '''

    def pre(c, r, level, cfg, t=t):
        i = {}
        print(c)
        ics = cfg['id_by_cls']
        hir = c._hir
        i['id'] = id = c._id
        if hir == 0:
            c._color = getattr(c, '_color', '#8bd124')
        i['itemTitleColor'] = getattr(c, '_color', '%s' % Color(hsl=color(hir)).hex)
        i['hir'] = c._hir
        i['level'] = c._level
        h = [('_id', c._id)]
        for k, v in simple_props(c):
            h.append((k, v))
        i['html'] = ',&nbsp;'.join(
            ['<span title="%s">%s</span>' % (k, v) for k, v in h]
        )
        ics[c] = id
        i['parent'] = 'null'
        p = getattr(c, '_parent', None)
        if p:
            i['parent'] = p._id
        i['type'] = c.type
        i['name'] = c.name
        i['title'] = '%(type)s %(name)s' % i
        i['description'] = '%(hir)s %(name)s' % i
        i['image'] = 'images/letters/%s.png' % c.type[0].lower()
        if hir > 0:
            i['templateName'] = 'axc'
        r.append(i)

    cfg['id_by_cls'] = {}
    cfg['id'] = -1
    res = walk_tree(root, pre, None, has_parent, cfg=cfg)
    # res[-1] = res[-1].rsplit(',', 1)[0]
    return ['var items = %s;' % res]


def to_ctree(root, **cfg):
    cfg['props_ign'] = cfg.get('props_ign', ())

    def pre(c, r, level, cfg):
        r.append('%s<li><a href="#">%s %s</a><ul>' % (level * ' ', c.type, c.name))

    def post(c, r, level, cfg):
        r.append('%s</ul></li>' % (level * ' '))

    res = walk_tree(root, pre, post, has_parent, cfg=cfg)
    return [r for r in res if r.strip()]


def to_dict(root, res=None, **cfg):
    if res is None:
        res = OD()
    for k, v in reversed(props(root)):
        if not type(v) == TypeMetaClass:
            # if k in ('alias', 'type', 'descr'):
            res[k] = v
        else:
            res[k] = to_dict(v, **cfg)
    return res


def to_tuples(root, **cfg):
    def pre(c, r, level, cfg):
        key = '.'.join(full_id(c))
        va = []
        for k, v in props(c):
            if not type(v) == TypeMetaClass:
                va.append([k, v])
        r.append([key, va])

    res = []
    walk_tree(root, pre, None, match=has_parent, res=res, cfg=cfg)
    return res


def to_class_tree(root, **cfg):
    def pre(c, r, level, cfg):
        ind = '   '
        n, t = c._cls, c.type
        r.append('%sclass %s(%s):' % (ind * (level - 1), n, t))
        for k in dir(c):
            if k.startswith('_'):
                continue
            v = getattr(c, k)
            if isinstance(v, simple_types):
                apo = "'''" if isinstance(v, str) else ''
                r.append('%s%s = %s%s%s' % (ind * level, k, apo, v, apo))

    res = walk_tree(root, pre, None, match=has_parent, cfg=cfg)
    res = [r for r in res if not 'tree_builder' in r]
    cfg['into'] = ('build_type_hirarchy', 'build_tree')
    cfg['skip_line_with'] = 'tree_builder'
    cfg['replace'] = (('(T, ', '('),)
    return res


def html_props(c, cfg):
    p = []
    if not cfg.get('no_props'):
        for k, v in props(c):
            if k in cfg['props_ign'] or not isinstance(v, simple_types):
                continue
            p += ['<tr><td>%s:</td><td><b>%s</b></td></tr>' % (k, v)]
    if p:
        p.insert(0, '<table>')
        p.append('</table>')
    return '\n'.join(p)


def to_html(root, **cfg):
    # ignored:
    hirarchy = root._hirarchy_list
    cfg['props_ign'] = cfg.get('props_ign', ())
    cc = cfg['css_clses'] = []

    def pre(c, r, level, cfg):
        cc = cfg['css_clses']
        if not c.type in cc:
            cc.append(c.type)
        classes = ()
        v = getattr(c, 'criticality', None)
        if v == 1:
            classes += ('critical',)
        if v == 2:
            classes += ('warning',)
        classes = ' '.join(classes)

        r.append('%s<%s class="%s">' % (level * '  ', c.type, classes))
        n = c._cls
        r.append('%s<b>%s</b> %s' % ((level + 1) * '  ', c._cls, c.descr))
        r.append(html_props(c, cfg))

    def post(c, r, level, cfg):
        r.append('%s</%s>' % (level * '  ', c.type))

    res = walk_tree(root, pre, post, has_parent, cfg=cfg)

    # setup css, using less:
    css = [
        ', '.join([f for f in [t.type for t in hirarchy]])
        + ''' {
        display:      inline-block;
        border-style: solid;
        border-width: 3px;
        padding:      5px; margin: 10px;
        }
    .critical {border-color: red}
    .warning {border-color: yellow}
    '''
    ]

    base = {'hue': 10, 'sat': 80, 'lum': 60, 'sat_bg': 30, 'lum_bg': 90}
    m = dict(base)
    hue_add = 70
    col = 'color: hsl(%(hue)s, %(sat)s%%, %(lum)s%%)'
    col_bg = 'color: hsl(%(hue)s, %(sat_bg)s%%, %(lum_bg)s%%)'

    for t in [t.type for t in hirarchy]:
        m['type'] = t
        m['col'] = col % m
        m['col_bg'] = col_bg % m

        css += ['%(type)s:before {content: "%(type)s:" attr(t);%(col)s}' % m]
        css += [
            ''.join(['%(type)s {background-%(col_bg)s; ' % m, 'border-%(col)s}' % m,])
        ]
        m['hue'] += hue_add

    res = '\n'.join([r for r in res if r.strip()])
    # res = res.replace('Service', 'service')
    res = [
        '<html>',
        '<body>',
        '<style>',
        '\n'.join(css),
        '</style>',
        '<div><table><tr><td>',
        '<div>',
        res,
        '</div>',
        '</td></tr></table></div>',
        '</body>',
    ]
    return res


def print_out(
    res, tmplfile=None, into='', outfile=None, skip_line_with=None, replace=(), **cfg
):
    """
    into: two seperators within a file where we put the result, keeping the seps,
          for a subsequent update (good for embedding e.g. into a md doc)
          string or tuple for same sep.
    """
    if tmplfile and not outfile:
        # -> tree_tmpl.html -> tree.html
        outfile = tmplfile.replace('_tmpl', '')
    out, tmpl = '\n'.join(res), ''
    if tmplfile:
        if not exists(tmplfile):
            raise Exception('template file not found', tmplfile)
        with open(tmplfile, 'r') as fd:
            tmpl = fd.read()

    pre, post = '', ''
    if into:
        if isinstance(into, str):
            into = (into, into)
        if not into[0] in tmpl or not into[1] in tmpl:
            # put at beginning then:
            tmpl = '\n%s\n%s\n%s' % (into[0], into[1], tmpl)
        pre, old = ('\n' + tmpl).split('\n' + into[0], 1)[:2]
        post = ''
        if '\n' in old:
            old = old.split('\n', 1)[1]
            old, post = old.split(into[1])[:2]
            post = post.split('\n', 1)[1]

    if skip_line_with:

        def sk(s):
            return '\n'.join([r for r in s.splitlines() if not skip_line_with in r])

        pre, post = sk(pre), sk(post)

    if pre or post:
        if outfile != file:
            # no seps in the target if target not the file
            into = ('', '')
        out = '\n'.join((pre, into[0], out, into[1], post)).strip()

    for k, v in replace:
        out = out.replace(k, v)

    if outfile:
        with open(outfile, 'w') as fd:
            fd.write(out)
    # print (out)
    return out



import re

from theming.unicode_chars import UCs


class BPFormat(object):
    """ bpython (curses?) format strings.
    Note: if it turns out they are curses complient then subclass.

    \x01 represents a colour marker, which
    can be proceded by one or two of
    the following letters:
    k, r, g, y, b, m, c, w, d
    Which represent:
    blacK, Red, Green, Yellow, Blue, Magenta,
    Cyan, White, Default
    e.g. \x01y for yellow,
         \x01G   Bold green
         \x01gb for green on blue background
         \x01gI  Inverse

    \x03 represents the start of the actual
        text that is output (in this case it's
        a %s for substitution)

    \x04 represents the end of the string; this is
        necessary because the strings are all joined
        together at the end so the parser needs them
        as delimeters
    """

    csi = '\x03'
    reset = '\x04'
    # K comes out as green strong on my term. have to investigate:
    color_map = {
        'green_strong': 'G',
        'black': 'k',
        'red': 'r',
        'green': 'g',
        'yellow': 'y',
        'blue': 'b',
        'magenta': 'm',
        'cyan': 'c',
        'white': 'w',
    }

    @staticmethod
    def format(s, col, bg=None, bold=False, inverse=False, reset=True):
        """ format('foo', 'green')"""
        r = BPFormat.reset
        if not reset:
            r = ''
        if len(col) == 2:
            # gI
            col, bg = col
        if col == col.upper():
            # GREEN
            col = col.lower()
            bold = 1
        elif not len(col) == 1:
            # green
            col = BPFormat.color_map[col]
        if bold:
            col = col.upper()
        if bg:
            if len(bg) != 1:
                col += BPFormat.color_map[bg]
            else:
                col += bg
        else:
            if inverse:
                col += 'I'
        res = '\x04\x01%s\x03%s%s' % (col, s, r)
        return res

    # --------------------------- any process can overwrite these to his liking:
    # debug, info, important, warn, alarm:
    cols = ['b', 'c', 'G', 'GI', 'RI']

    @staticmethod
    def fmt_debug(s):
        return BPFormat.format(s, BPFormat.cols[0])

    @staticmethod
    def fmt_info(s):
        return BPFormat.format(s, BPFormat.cols[1])

    @staticmethod
    def fmt_important(s):
        return BPFormat.format(s, BPFormat.cols[2])

    @staticmethod
    def fmt_warn(s):
        return BPFormat.format(s, BPFormat.cols[3])

    @staticmethod
    def fmt_alarm(s):
        return BPFormat.format(s, BPFormat.cols[4])

    # can be set by processes, like config mgmt for [D]...[S]... matches:
    #  {re.compile(r'(\[D\]\w*)'): 'info'} -> show all "foo [D]bar baz" in info color
    rules = {}
    # if order is wanted:
    rules_list = []
    for table_char in UCs.VU, UCs.JUNC, UCs.H, UCs.GOOD:
        rules_list.append([re.compile(r'(%s)' % table_char), 'g'])
    rules_list.append([re.compile(r'(%s)' % UCs.BAD), 'R'])

    @staticmethod
    def fmt(s):
        # first the list, order here, map rules than won't match:
        F = BPFormat
        l = [F.rules_list, F.rules.items()]
        for p in l:
            for match, rule in p:
                func = getattr(F, 'fmt_%s' % rule, 0)
                if func:
                    # rule is like 'info'
                    s = match.sub(func('\\1'), s)
                else:
                    # rule is a color:
                    s = match.sub(F.format('\\1', rule), s)
        return s


if __name__ == '__main__':
    # run this in bpython to see, normal shell should show unchanges output:
    BPFormat.rules = {re.compile(r'(\[D\]\w*)'): 'info'}
    print(BPFormat.fmt('foo [D]daemonname[S]other\nbar [D]daemon2'))



#!/usr/bin/python -tt
from __future__ import absolute_import, division, print_function

import unittest
from functools import partial

import structlog as sl

# import ax.rx

try:
    pass

    have_gevent = True
except:
    have_gevent = False


# great to not have to inspect log files for the result:
RL = sl.ReturnLogger
# logger factory:
fac = lambda L, *a, **kw: L(*a, **kw)


def cfg_insert_proc(_, __, ev, **cfg):
    ev.update(cfg)
    return ev


class TestStructlogging(unittest.TestCase):
    def setUp(self):
        """
        """
        self.p1 = partial(cfg_insert_proc, p1='p1')
        self.p2 = partial(cfg_insert_proc, p2='p2')
        self.p3 = lambda _, __, ev: str(ev)

    def tearDown(self):
        """
        """

    def test_basic1(self):
        l = sl.wrap_logger(RL(), processors=[self.p1, self.p2])
        res = l.info('foo')[1]
        self.assertIn('p1', res)
        self.assertIn('p2', res)
        self.assertEqual(res['event'], 'foo')
        l2 = l.bind(key1='val1')
        res = l.debug('foo2')
        self.assertNotIn('key1', res)
        res2 = l2.debug('foo2')
        self.assertNotIn('key1', res2)

    def test_configure(self):
        l = sl.get_logger(RL(), processors=[self.p1, self.p2, self.p3])
        res = l.info('p1')
        # pdb.set_trace()


if __name__ == '__main__':
    unittest.main()



#  to not break import theming.html, after a Py3 compat induced
# rename of the html.py module into html_tools.py:
import sys


def func_name_of_callable(
    f, key='__name__' if sys.version_info[0] < 3 else '__qualname__'
):
    """Returns the name of a function, e.g. for logs so that the reader knows
    what is being executed, w/o clutter

    e.g. str(f) = "<bound method AXConfiguratorMain.handle_termination_signal of <Products.WPSCommon.TR69.AXConfigurator.AXConfiguratorMain object at 0x7f1a1cfe7f10>>"
    is hard to read while not all callables have .__name__
    while we return just "handle_termination_signal" or "AXConf.handle.." in v3

    Works also for nested partials (but not anonymous lambdas)
    """
    # todo: decorators / wrapped, classname of __call__, ...
    f1 = f
    while hasattr(f1, 'func'):
        f1 = f1.func
    return getattr(f1, key, str(f1))



"""

sys.argv: 
    build/bin/<run_script> --log_fmt=10 foo bar -baz

Remember: run_script func-calls our def app_run(), with these as sys.args

if we can import $app (checked in loader) -> we check app_mod for 'foo' function
if present we call it right away, via run_app(foo)

if not app_mod or not presnt than this is a subproc call, e.g. for bash.

-> we parse all flags before "foo" in devapps then Popen the rest

ALL FLAGS BEFORE THE APP FUNCTION HAVE TO BE GIVEN WITH =!

WRONG: --log_level info
RIGHT: --log_level=info

(otherwise we'll try to lookup 'info' in the app module)
"""
import json

# pylint: disable=E0211
import os
import shutil
import signal
import subprocess as sp
import sys
import time
from functools import partial
from inspect import _empty, signature

from devapp import load
from devapp.app import FLG, app, flag, run_app
from devapp.spec import tools as spec_tools
from devapp.tools import has_tty, write_file

try:
    import psutil
except ImportError as ex:
    psutil = None


exists = os.path.exists
FS = spec_tools.FS
pyenv = lambda k, dflt=None: load.py_env.get(k, dflt)
proc_argv = []
env = os.environ

flag.bool('show', False, 'Dry run, only show args')
m = 'Additional $PATH, added to :%s'
flag.string('add_path', '', (m % env['PATH']).replace(':', '\n-'))

# always good to have (for md renders and the like):
if not env.get('term_with'):
    try:
        c, r = os.get_terminal_size(0)
    except Exception as ex:
        c, r = 80, 25
    env['term_width'], env['term_height'] = str(c), str(r)

# prevent problems when user exported it with / at the end:
_ = 'DA_DIR'
if _ in env:
    while env[_].endswith('/'):
        env[_] = env[_][:-1]


class DAFunctions:
    """
    Functions accessible via devapps' `da` utility
    """

    @staticmethod
    def _da_dir(die=True):
        da_dir = env.get('DA_DIR') or []
        # len for security, we rm -rf below
        if not da_dir or len(da_dir) < 2:
            msg = 'No $DA_DIR. Provide --da_dir or --da_name'
            app.error(msg)
            sys.exit(0)
        if not da_dir:
            if die:
                sys.exit(1)
        return da_dir

    @staticmethod
    def status():
        """Infos about the devapps installation"""
        return {'status': 'installed'}

    class tools:
        """general tools"""

    class spec:
        @staticmethod
        def _fn():
            return DAFunctions._da_dir() + '/conf/SPEC.py'

        @staticmethod
        def edit():
            if not has_tty:
                app.die('Require tty')
            vim = env.get('EDITOR') or 'vim'
            spec = DAFunctions.spec._fn()
            os.system(vim + ' ' + spec)

        @staticmethod
        def dict(tree=True, src=False, links=False):
            """Prints a serialized version of the spec to stdout
            The spec is expected in $DA_DIR/conf/SPEC.py"""
            spec = DAFunctions.spec._fn()
            g = DAFunctions.tools.exec_file(spec, get_src=True)
            r = g['root']
            m = {}
            if tree:
                m['tree'] = g['to_dict'](r)
            if links:
                m['links'] = g['dump_links'](r)
            if src:
                m['src'] = g['source']
            return m

        @staticmethod
        def apps(run=None):
            """ 
            Printing all app starter files.  

            The `run` argument allows for further processing (e.g. da spec.apps sc.status).
            I.e. `run` is convenience, for a loop like 
            `while read -r n; do $n sc.status; done < <(da -n wifi_single spec.apps | jq -r .[] )`
            
            Example: da spec.apps sc.status
            """
            # TODO: understand run as a list, i.e. pass a full command with args
            da_dir = DAFunctions._da_dir(die=False)
            db = da_dir + '/build/bin'
            if not os.path.exists(db):
                app.error('Empty - no build', dir=db)
                return []
            apps = [db + '/' + d for d in os.listdir(db)]
            if not run:
                return apps
            else:
                for app_ in apps:
                    app.info(app_, running=run)
                    run_proc([app_, run])

        @classmethod
        def clean(cls, confirm=False, conda_envs=False, etc=False, keep_running=False):
            """
            Deletes everything related to current service which is built at spec build runs
            except conda_env, unless envs=True is given.

            This is useful for tests.
            """
            if not has_tty and not confirm:
                app.error('Have no tty and force not set - ignoring')
            D = env['DA_DIR']

            def rmd(d, do=False):
                if exists(d):
                    if do:
                        app.info('Removing', dir=d)
                        shutil.rmtree(d)
                    else:
                        app.info('Keeping', dir=d)

            rmd(D + '/build', True)
            # [ # we don't have that often:
            #    run_proc([s, 'devapp.clean', 'keep_running=%s' % keep_running])
            #    for s in cls.apps()
            # ]
            if len(D) > 4:
                for sw, d in (conda_envs, 'envs'), (etc, 'etc'):
                    rmd(D + '/' + d, sw)


# alias
DAFunctions.apps = DAFunctions.spec.apps


def verify_lib(lib):
    app.die('Missing library', lib=lib) if not lib in globals() else 0


def psutil_str_to_dict(v):
    s = str(v).replace('"', '').replace("'", '')

    try:
        kvs = s.split('(', 1)[1][:-1].split(', ')
        kvs = [kv.split('=', 1) for kv in kvs]
        return dict([(k, auto_type(v)) for k, v in kvs])
    except:
        return {}


def dump_proc(p, **kw):
    pu2s = psutil_str_to_dict
    m = {
        'pid': p.pid,
        'children_count': len(p.children()),
        '_pretty_': pu2s(p),
    }
    pa = [
        'cmdline',
        'connections',
        'cpu_affinity',
        'cpu_num',
        'cpu_percent',
        'cpu_times',
        'create_time',
        'cwd',
        'exe',
        'gids',
        'io_counters',
        'is_running',
        'memory_full_info',
        #'memory_info',
        #'memory_info_ex',
        'memory_percent',
        'name',
        'nice',
        'num_ctx_switches',
        'num_fds',
        'num_threads',
        'open_files',
        'parent',
        #'parents',
        'ppid',
        'status',
        'terminal',
        'threads',
        'uids',
        'username',
    ]
    subs = [
        'cpu_times',
        'gids',
        'io_counters',
        'memory_full_info',
        'num_ctx_switches',
        'parent',
        'uids',
    ]
    filter = kw.get('filter', '')
    if filter:
        pa = [k for k in pa if filter in k]
    if kw.get('env') or 'env' in filter:
        pa.append('environ')
    if kw.get('memory_maps') or 'memory_maps' in filter:
        pa.append('memory_maps')
    n = dict([(k, getattr(p, k)()) for k in pa])
    for s in subs:
        v = n.get(s)
        if v:
            n[s] = pu2s(v)
    for fn in 'threads', 'open_files':
        n[fn] = [pu2s(t) for t in n.get(fn, [])]

    ns = n.pop('connections', 0)
    if ns:
        try:
            n['conn'] = [{'l': pu2s(c.laddr), 'r': pu2s(c.raddr)} for c in ns]
        except Exception as ex:
            n['conn'] = []

    m.update(n)
    tree = kw.get('tree', 0)
    if tree:
        kw['tree'] = tree - 1
        m['c'] = [dump_proc(pc, **kw) for pc in p.children()]
    return m


class AppFunctions:
    """The app API, within a daemon env"""

    class systemctl:
        def __init__(self, m, *a):
            getattr(self.__class__, m)(*a)

        @staticmethod
        def cmd(_cmd):
            """called from to stop, start, ..."""
            l = ['systemctl', '--no-pager', _cmd, pyenv('name_unit')]
            l.insert(1, '--user') if os.geteuid() else ''
            return run_proc(l)

    sc = systemctl

    class os:
        @classmethod
        def _proc(cls, all=False):
            # we do not need to rely on the pid file but have the process name:
            # 16 bytes proces names in unix
            verify_lib('psutil')
            n = AppFunctions.os.name()
            # we do not use the pid file but the process name:
            ps = [
                proc
                for proc in psutil.process_iter()
                if proc.name() == n and 'start' in proc.cmdline()
            ]
            if all:
                return ps
            # e.g. '/home/gk/devapps/test/wifi_single/run/AXWiFi.Hub'
            n = pyenv('app_run_exe_link')
            # loop thru ALL processes and find the one matching exe_link and 'start':
            return [pi for pi in ps if n in pi.cmdline()]

            return ps

        @classmethod
        def proc(
            cls,
            pid=-1,
            env=False,
            memory_maps=False,
            filter='',
            tree=0,
            parent=False,
            all=False,
        ):
            """Extensive process info

            pid: Main app process if not given
            env: Include process environ
            memory_maps: Include memory_maps (loaded libs)
            filter: filter on substring
            tree: Include children down to this level 
            
            """
            l = dict(locals())
            ps = [psutil.Process(pid)] if pid > -1 else cls._proc(all=all)
            if parent:
                ps = [psutil.Process(p.parent().pid) for p in ps]

            return [dump_proc(p, **l) for p in ps]

        @classmethod
        def name(cls):
            """"""
            # e.g. hub.axwifi
            # 16 bytes proces names in unix
            return pyenv('app_run_script')[:15]

        @classmethod
        def pid(cls, all=False):
            """
            Returns pid of process matching $app_run_exe_link and 'start'
            
            all: Return ALL started processes with our name, in a list
            (started in any DA_DIR, within the pid namespace)
            """
            ps = AppFunctions.os._proc(all=all)
            if not all:
                return ps[0].pid if ps else None
            else:
                return [pi.pid for pi in ps]

        @classmethod
        def signal(cls, sig=int(signal.SIGTERM), ignore_not_found=False, all=False):
            """Send a signal to the running process"""
            sig = int(sig)
            ps = AppFunctions.os._proc(all=all)
            if ps:
                pids = [p.pid for p in ps]
                app.info('Sending signal', pid=pids, signal=sig)
                [os.kill(p, sig) for p in pids]
            else:
                if not ignore_not_found:
                    app.error('Process not found')

        @classmethod
        def poll(cls, freq=100, max=60000, on_success=None):
            p, t1, maxm = None, time.time(), max / 1000.0
            while time.time() < (t1 + maxm):
                p = cls.pid()
                if p:
                    break
                time.sleep(freq / 1000.0)
            if not p:
                sys.exit(1)
            if on_success:
                app.info('Running on_success handler', cmd=on_success)
                if os.system(on_success):
                    app.error('on_success handler failed', cmd=on_success)
                    sys.exit(1)
            return p

    class devapp:
        @staticmethod
        def env():
            return load.py_env

        @staticmethod
        def fs():
            with open(pyenv('build_dir') + '/fs_stack.json') as fd:
                return json.loads(fd.read())

        @staticmethod
        def links():
            with open(pyenv('build_dir') + '/links.json') as fd:
                return json.loads(fd.read())

        @staticmethod
        def clean(keep_running=False):
            un = pyenv('name_unit')
            if not keep_running:
                AppFunctions.systemctl.stop()
                AppFunctions.stop(ignore_not_found=True)
            for k in 'build', 'var':
                bd = pyenv(k + '_dir')
                if os.path.exists(bd) and len(bd.split('/')) > 3:
                    app.info('Removing directory', dir=bd)
                    run_proc(['rm', '-rf', bd])
            sd = units_dir() + '/' + un
            if os.path.exists(sd):
                app.info('Removing unit', unit=sd)
                os.unlink(sd)

    @staticmethod
    def status(short=True):
        p = AppFunctions.os._proc()
        app.info('Running') if p else app.die('Not running')

    # alias
    d = devapp


# sc.start, sc.stop:..
AF = AppFunctions
[
    setattr(AF.systemctl, k, partial(AF.sc.cmd, _cmd=k))
    for k in ('stop', 'start', 'restart', 'status', 'reload')
]

# reload, stop:.. (start is always mapped within the app module or to app_start)
[
    setattr(AF, k, partial(AF.os.signal, sig=sig))
    for k, sig in (
        ('stop', signal.SIGTERM),
        ('quit', signal.SIGQUIT),
        ('kill', signal.SIGQUIT),
        ('reload', signal.SIGHUP),
    )
]


def units_dir():
    return (
        os.environ['HOME'] + '/.config/systemd/user'
        if os.geteuid()
        else '/etc/systemd/system'
    )


def unit_files(da_name):
    d = units_dir()
    return [d + '/' + i for i in os.listdir(d) if '-%s.service' % da_name in i]


def da_name():
    return os.environ['DA_DIR'].rsplit('/', 1)[-1]


def set_da_dir_by_da_name(name, nofail=False):
    """This is questionable: Attempting to provide the convenient -n wifi to da util
    and rely on it deriving the right DA_DIR
    We had the option to write a $HOME/.local/devapps file at spec build mapping names to dirs
    but what if the spec is built elsewhere?
    So for now we look into the unit files.
    # todo: won't work for unit less utility builds though, then we need the mapping file for this

    Called by the CLI parser before run_app, i.e. we don't have app logging yet.
    """
    units = unit_files(name)
    if not units:
        if nofail:
            return None
        raise Exception('Cannot derive DA_DIR - no units')

    with open(units[0]) as fd:
        s = fd.read()

    env['DA_DIR'] = dd = (
        s.split('\nExecStart ', 1)[1].split('/build')[0].rsplit(' ', 1)[-1]
    )
    return dd


def import_and_expose_da_level_modules():
    """ import of modules only for da utility, not for apps"""
    from devapp.spec import build as sb
    from devapp.tools import exec_file

    DAFunctions.spec.build = sb.build
    DAFunctions.tools.exec_file = exec_file


def devapps_run():
    """
    The 'da' utility, wich is controlling the solution, not an app

    From setup.py: 'da=devapp.run:devapps_run',
    """
    import_and_expose_da_level_modules()

    flag.string('da_dir', '$DA_DIR', 'Solution directory', short_name='d')
    flag.string('da_name', '$DA_NAME', 'Solution name', short_name='n')
    args = sys.argv
    parse_flags_funcname_and_args(args, for_devapps=True)
    if not 'DA_DIR' in env:
        set_da_dir_and_name_from_current_directory()
    # special case: Help required:
    # the or is for: da -n wifi_single -h
    if is_help_intent() or RUN['func_name'] == 'bash' and not RUN['func_args']:
        f = DAFunctions
    else:
        f = get_func(DAFunctions)
    if callable(f):
        f = with_args(f)
        # sys.argv = RUN['argv']
        return run_app(f, wrapper=wrapped_app_func)
    # he called a command line program:
    f = partial(run_exit, parametrize_external_process())
    run_app(f, wrapper=partial(wrapped_app_func, ext_proc=True))


is_help_intent = lambda: len(sys.argv) == 2 and sys.argv[-1] in [
    '-h',
    '-hh',
    '--help',
    '-hhh',
    '--helpfull',
]


def wrapped_app_func(f, ext_proc=False):
    """apps have the PATH in their environ. Only via this flag we can allow
    additional PATH entries within apps processes
    """
    if FLG.show:
        if not ext_proc:
            from theming.absl_color_help import call_doc

            call_doc(f, level=3, render=True)
            return

    if FLG.add_path:
        have = env['PATH'].split(':')
        [have.insert(0, i) for i in reversed(FLG.add_path.split(':')) if not i in have]
        env['PATH'] = ':'.join(have)
    return f()


def add_app_functions_for_usage_help():
    """-h lists all app functions - we want these here as well:"""
    procs = []
    for k, v in pyenv('functions', {}).items():
        # convention: {'foo': ['a', 'b']} -> a is a process ,b param for it - python functions
        # will be anyway in the list:
        if isinstance(v, list):
            f = partial(app_run_exit, v)
            setattr(AppFunctions, k, partial(app_run_exit, v))
            procs.append(k)
    return procs


def app_run(args):
    """
    The app_run scripts in build/bin call this - from python

    - Unsharing done outside of us
    - FS is built (we dropped building at start up time - not yet sure regarding unshared changes, os.system unshare -> FS run, recall from caller

    Find what to run:
    e.g. for sys.argv = '--log_level=10 start':

    - We do not support supplying an app.module on the cli, i.e. for devapp all cmds must be in app module as funcs
    - We do support a nested api within the app module though

    """
    # func_names: 'start', 'metrics.get', 'stop', 'redis-server'
    # dash flags before and after
    # procs = add_app_functions()
    # this builds the aliases:

    # TODO: This is a mess and needs rework when the app is NOT a function - parametrize_external_process should not be needed
    # or be simpler, i.e. this must do that work already:
    parse_flags_funcname_and_args(args)
    func_name = RUN['func_name']

    ApplicationFunctions = AppFunctions

    if load.app_mod:

        if func_name == 'start':
            # app was importable, for devapps run['start'] = '<modulename>' -> ignored:
            add_prod_flagfile()
            register_signals(proc_signal_handler)
            return run_app(load.app_mod.start, wrapper=wrapped_app_func)

        appns = getattr(load.app_mod, 'Functions', None)
        if appns:

            class ApplicationFunctions(appns, AppFunctions):
                pass

    f = get_func(ApplicationFunctions)

    if is_help_intent():
        add_app_functions_for_usage_help()
        f = ApplicationFunctions

    if callable(f):
        f = with_args(f)
        add_prod_flagfile()
        return run_app(f, wrapper=wrapped_app_func)

    # this is a subproc run. Lets run it under an initiazlied dev_app proc, i.e. get
    # the flags:
    f = partial(app_run_exit, parametrize_external_process())
    run_app(f, wrapper=partial(wrapped_app_func, ext_proc=True))


pidfile = lambda: env['pidfile']
pidstarter = lambda: env['var_dir'] + '/pidfile.starter'


def add_prod_flagfile():
    pf = pyenv('etc_dir') + '/prod.flags'
    if exists(pf) and '--flagfile' not in sys.argv:
        sys.argv.append('--flagfile')
        sys.argv.append(pf)


def write_pid(sp_pid=None, remove=False):
    # is given for spawned ones else its ours:
    pidst = os.getpid()
    # for devapps both are ident:
    for fn, p in (pidfile(), sp_pid), (pidstarter(), pidst):
        if remove:
            if os.path.exists(fn):
                os.remove(fn)
        else:
            write_file(fn, str(p) + '\n', mkdir=True)


def app_run_exit(args):
    return run_exit(args, for_app=True)


def run_exit(args, for_app=False):
    """Run a subprocess"""
    # TODO: basic support for && and pipes
    # User called another function or process but we do warn him about that:
    # only now we can report this:
    if not for_app and os.path.exists(FLG.da_dir):
        app.debug('cd', dir=FLG.da_dir)
        os.chdir(FLG.da_dir)

    if load.app_import_err:
        app.error('App import error', exc=load.app_import_err)

    sys.exit(run_proc(args, for_app=for_app, handle_sigs=True))


def run_helper(args):
    """args can refer to inner commands, in order to gather values"""
    proc = sp.Popen(args, stdout=sp.PIPE, stderr=sp.PIPE)
    res = proc.communicate()
    if proc.returncode != 0:
        app.die('Failing', args=args, out=res[0], err=res[1])
    r = res[0].decode('utf-8')
    if r and r[-1] == '\n':
        r = r[:-1]
    return r


def run_proc(args, for_app=False, handle_sigs=False):
    args = [run_helper(a) if isinstance(a, list) else a for a in args]
    args = [str(l) for l in args]
    if FLG.show:
        s = lambda i: '"%s"' % i if not i.startswith('-') else i
        app.info('Dry Run', cmdline=' '.join([s(a) for a in args]))
        return 0
    # raises:
    try:
        proc = sp.Popen(args)
    except Exception as ex:
        app.error(
            'Command run error',
            cmd=args[0],
            args=args[1:],
            err=ex,
            pid_starter=os.getpid(),
        )
        raise
    app.debug(args[0], args=args[1:], pid=proc.pid, pid_starter=os.getpid())
    if handle_sigs:
        # for app:
        register_signals(partial(proc_signal_handler, pid=proc.pid))

    proc.communicate()
    return proc.returncode


# SIGKILL we would not see, anyway:
termsigs = signal.SIGINT, signal.SIGTERM, signal.SIGQUIT

app_sig_catchers = {
    signal.SIGHUP: 'reload',
    signal.SIGQUIT: ['quit', 'stop'],
    signal.SIGTERM: 'stop',
    signal.SIGINT: ['interrupt', 'stop'],
}


def get_handler(mod, catcher):
    if not catcher:
        return
    if isinstance(catcher, str):
        return getattr(mod, catcher, None)
    return getattr(mod, catcher[0], None) or getattr(mod, catcher[1], None)


def proc_signal_handler(sig, frame, pid=None):
    if load.app_mod:
        c = app_sig_catchers.get(sig)
        f = get_handler(load.app_mod, c)
        if f:
            fn = getattr(f, '__qualname__', f)
            app.warn('Calling %s' % fn, signal=sig, func=fn)
            f()
        else:
            app.warn('No signal handler for %s' % str(c), signal=sig)

    if pid:
        try:
            # kill -QUIT <this master process pid> should -9 the child:
            if sig == signal.SIGQUIT:
                app.warn('KILLING HARD (SIGKILL)', pid=pid)
                os.kill(pid, signal.SIGKILL)
            else:
                # forwarding:
                app.warn('Signal', signal=sig, pid=pid)
                os.kill(pid, sig)
        except Exception as ex:
            app.error('Could not send signal', pid=pid, signal=sig, exc=ex)
    if sig in termsigs:
        if pid:
            write_pid(pid, remove=True)
        sys.exit(1)


def register_signals(handler):
    # signal.SIGTSTP is forwarded to the child (tested: redis)
    [signal.signal(s, handler) for s in termsigs]
    signal.signal(signal.SIGHUP, handler)


# process = [proc for proc in psutil.process_iter() if proc.name == "YourProcess.exe"].


if __name__ == '__main__':
    app_run(args=sys.argv)


# cheap function finder and sig mapper.
# TODO: re-use the stuff from axc2, with parametrz.
# classes and recursive defaults.
# attrs or dataclasses ?

RUN = {'argv_orig': [], 'argv': [], 'func_args': [], 'func_name': None}


def parametrize_external_process():
    """ External process -> all flags before "func_name" (which is the process) belong to flags, 
    rest is process args, e.g. --log_level debug tree -L 2"""
    argv = RUN['argv_orig']
    pos = argv.index(RUN.get('func_name_aliased') or RUN['func_name'])
    sys.argv.clear()
    sys.argv.extend(argv[:pos])
    return argv[pos:]


def parse_flags_funcname_and_args(sys_argv, for_devapps=False):
    """Fills the RUN structure """

    def get_all_flags():
        from absl import app

        m = app.FLAGS.flags_by_module_dict()
        l = {}
        for k, v in m.items():
            for f in v:
                l['--%s' % f.name] = f
                l['-%s' % f.short_name] = f
        return l

    fn = None
    RUN['argv_orig'].extend(sys_argv)
    defined_flags = get_all_flags()
    RUN['argv'].append(sys_argv[0])  # the main program (this)
    l = sys_argv[1:]
    while l:
        a = l.pop(0)
        if a.startswith('-'):
            if for_devapps:
                set_da_dir_and_name_from_cli_param(a, args=l)
            RUN['argv'].append(a)
            f = defined_flags.get(a)
            if f and not 'Bool' in str(f):
                if not l:
                    raise Exception('Value required for %s' % a)
                RUN['argv'].append(l.pop(0))
        else:
            if not fn:
                aliased = pyenv('functions', {}).get(a)
                if aliased:
                    av = RUN['argv_orig']
                    avo = list(av)
                    av.clear()
                    av.extend(avo[: len(RUN['argv'])])
                    av.extend(aliased)
                    av.extend(l)
                    RUN['func_name_aliased'] = aliased[0]
                RUN['func_name'] = fn = a
            else:
                RUN['func_args'].append(a)
    if not fn:
        RUN['func_name'] = 'bash'
        # add the da path into the environ if not in yet for the bash env:
        RUN['argv_orig'] = RUN['argv'] + [
            'bash',
            '--rcfile',
            '%s/shell/activate' % devapps_dir(),
        ]


def _cli_val(a, args):
    return a.split('=', 1)[1] if '=' in a else args[0]


def set_da_dir_and_name_from_current_directory():
    p = os.getcwd()
    while len(p) > 2:
        if exists(p + '/conf/SPEC.py'):
            env['DA_DIR'] = p
            return
        p = p.rsplit('/', 1)[0]


def set_da_dir_and_name_from_cli_param(a, args):
    # Already before run_app is called we export DA_DIR and DA_NAME
    # if given - since they are fundamental
    if a.startswith('--da_dir') or a == '-d':
        env['DA_DIR'] = _cli_val(a, args)
        env['DA_NAME'] = env['DA_DIR'].rsplit('/', 1)[-1]
    elif a.startswith('--da_name') or a == '-n':
        env['DA_NAME'] = _cli_val(a, args)
        if not 'DA_DIR' in env:
            set_da_dir_by_da_name(env['DA_NAME'], nofail=True)


def devapps_dir():
    """returning the directory of the devapps repo, i.e. where shell, python, ... directories are"""
    import structlog

    # we run da to get here, i.e. our env is base/envs/devapps -> structlog is within
    # Only restriction: the devapps env must be in default location of the base env
    # (and devapps always the repo version, which is in the base package)
    return structlog.__file__.rsplit('/envs/')[0] + '/repos/devapps'


def get_func(ns):
    fn = RUN['func_name']
    parts, h = fn.split('.'), []
    while parts:
        h.append(parts.pop(0))
        ns = getattr(ns, h[-1], None)
        if not ns:
            return None if len(h) == 1 else '.'.join(h)
    return ns


def auto_type(v):
    for k, b in ('true', True), ('false', False):
        if str(v).lower() == k:
            return b
    for typ in int, float:
        try:
            return typ(v)
        except:
            pass
    if not v:
        return ''
    return str(v)


def with_args(f):
    """f is a function an we have its args -> return a parametrized version, ready for run_app"""
    # get command args:
    sig = signature(f).parameters
    sigv = list(sig.values())
    a, kw, have = (), {}, []

    def sig_type(v, P):
        if not P:
            return auto_type(v)
        elif P.default is None:
            return v
        elif P.default is _empty:
            return auto_type(v)
        # foo=bool (type w/o a value. mypy later. maybe)
        elif isinstance(P.default, type):
            return P.default(v)
        elif isinstance(P.default, bool):
            return str(v).lower() in ['1', 'true']
        else:
            try:
                return type(P.default)(v)
            except Exception as ex:
                print('breakpoint set')
                keep_ctx = True

    # argv:
    for p in RUN['func_args']:
        if '=' in p:
            k, v = p.split('=', 1)
            P = sig.get(k)
            kw[k] = sig_type(v, P)

        else:
            P = None
            while sigv and not P in have:
                P = sigv.pop(0)
                have.append(P)
            a += (sig_type(p, P),)
    # wrap command function with args:
    f = partial(f, *a, **kw)
    return f



#!/usr/bin/env python
"""
Tool to quickly build and app with plugins.

Usage: See lc.py

"""
import os
import sys
from fnmatch import fnmatch
from importlib import import_module

from devapp.tools import dir_of


def get_docstr(app=None):
    if not app:
        app = sys.argv[0]

    D = '''
Usage: %(app)s ACTION [--flag[=value] ...]

Available Plugins:
<PLUGINS>

Help:
    %(app)s <action> <-h|--helpfull> [match]

Note:
    - Action shortcuts understood, e.g. action "foo_bar_baz" = fbb
    - Plugins are taken on first found basis
    - Flags also have shortcut versions (e.g. -hf for --helpfull)

Example:
    %(app)s %(exmpl)s -hf log # all flags about logging
    '''
    D = D % {'app': os.path.basename(app), 'exmpl': example[0]}
    return D


class Plugins:
    pass


plugins = lambda: [p for p in dir(Plugins) if not p.startswith('_')]
_short = lambda p: ''.join([k[0] for k in p.split('_') if k])
shorts = lambda: {_short(k): k for k in plugins()}

example = ['']


def usage(e=0, app=None):
    ps = plugins()
    a = '\n'.join(['    - \x1b[1;32m%s\x1b[0m' % a for a in ps])
    if ps:
        example[0] = ps[0]
    print(get_docstr(app).replace('<PLUGINS>', a).replace(os.environ['HOME'], '~'))
    sys.exit(e)


helpflags = set(['-h', '--help', '--helpfull'])


def main(argv=None):
    """allows to pass argv[1:] (for testing)"""
    # if argv is not None:
    #     while len(sys.argv) > 1:
    #         sys.argv.pop()
    #     sys.argv.extend(argv)

    argv = sys.argv
    n_app = argv[0].rsplit('/', 1)[-1]  # e.g. app or ops
    find_plugins(n_app)
    if not plugins():
        msg = (
            'No plugins found. Create plugins/%s_<namespace>/ folder(s), '
            'containing importable python modules.'
        )
        print(msg % n_app)
        sys.exit(1)
    plugin = sys.argv[1] if len(sys.argv) > 1 else 'x'
    plugin = shorts().get(plugin, plugin)
    if not plugin in plugins():
        ec = 0 if any([a for a in sys.argv if a in helpflags]) else 1
        usage(ec)

    plug = plugname = getattr(Plugins, plugin)
    if isinstance(plug, str):
        plug = import_module(plug)

    sys_argv_rm_minus_h_for_hf(argv, plug, plugname)

    plug.main()


def sys_argv_rm_minus_h_for_hf(argv, plug, plugname):
    h, hf = False, False
    r = []
    for a in argv:
        if a in ('-h', '--help'):
            h = True
            continue
        if a in ('-hf', '--helpfull'):
            hf = True
        r.append(a)
    sys.argv.clear()
    sys.argv.extend(r)
    if h and not hf:
        sys.argv.append('-hf')
        sys.argv.append(plugname)
    if h or hf:
        # at -h, print out the module doc already here, since -h won't:
        # edit - no it actually does (doc pp -h)
        print()
        # print(plug.__doc__.strip())
        # print()


def all_plugin_dirs(name):
    """On name collisions we take first one found in sys.path
    => $PYTHONPATH overrules site-packages
    """
    all = {}

    def scan_sys_pth(d, all=all, exists=os.path.exists, name=name):
        for l in os.listdir(d):
            n = '%s/%s/plugins/%s_%s' % (d, l, name, l)
            if exists(n):
                name = os.path.basename(n)
                if not name in all:
                    all[name] = n

    [scan_sys_pth(d) for d in sys.path if os.path.isdir(d)]
    return all.values()


def find_plugins(n_app, pth='./plugins', match='*'):
    # ../devapp/bin/app.py' -> app:
    ds = all_plugin_dirs(n_app)
    [add_plugins(d, match) for d in ds]
    return main


def add_plugins(dp, match, _have=[]):
    dplugs = os.path.abspath(dp)
    d = os.path.dirname(dplugs)
    if not d in sys.path:
        sys.path.insert(0, dir_of(dplugs))
    dpre = dplugs.rsplit('/', 1)[-1]

    def c(fn, match=match):
        f = fn.split('.py')
        if len(f) == 2 and not fn.startswith('_'):
            if fnmatch(f[0], match):
                return f

    plugs = [fn for fn in os.listdir(dplugs) if c(fn)]
    for p in plugs:
        n = p.rsplit('.py')[0]
        # plug = import_module('%s.%s' % (dpre, n))
        plug = '%s.%s' % (dpre, n)
        setattr(Plugins, n, plug)


if __name__ == '__main__':
    main()



# adding builtins


def test_dict_tree():
    a = {'foo': {'bar': 'baz'}}
    d = DT(a)
    assert isinstance(a, dict)
    assert d.foo.bar == 'baz'
    assert d.foo['bar'] == 'baz'


def test_dict_tree_matched():
    a = {'foo': {'bar': 'baz'}}
    d = T(a)
    assert isinstance(a, dict)
    assert d.foo.bar == 'baz'
    assert d.foo['bar'] == 'baz'
    assert isinstance(a, dict)
    assert d.f.ba == 'baz'
    assert d.fo['bar'] == 'baz'



import os
from fnmatch import fnmatch

from devapp import FLG, app, flag, run_app

is_dir = os.path.isdir
exists = os.path.exists


flag.str('pkg', '', 'package to download and install in dev mode')


def get_package_source(pkg):
    pass


def get_meta_yaml(match, download_if_not_local=False):
    ad = app.pkgs_dir
    if not exists(app.pkgs_dir):
        raise app.die(
            'No pkgs dir.',
            dir=app.pkgs_dir,
            hint=' Maybe keep_pkgs was set to false when creating this in constructor',
        )
    ap = os.listdir(app.pkgs_dir)
    app.log.debug('Packages', count=len(ap))
    p = [p for p in ap if fnmatch(p, match) and is_dir(ad + '/%s' % p)]


def dev_on_package(match):
    ad = app.pkgs_dir
    if not exists(app.pkgs_dir):
        raise app.die(
            'No pkgs dir.',
            dir=app.pkgs_dir,
            hint=' Maybe keep_pkgs was set to false when creating this in constructor',
        )
    ap = os.listdir(app.pkgs_dir)
    app.log.debug('Packages', count=len(ap))
    p = [p for p in ap if fnmatch(p, match) and is_dir(ad + '/%s' % p)]
    for pkg in p:
        get_package_source(pkg)
        # dev_install(pkg)
    return p


def main():
    """Called from max d package"""
    return dev_on_package('*%s*' % FLG.pkg)


install_dev_pkg = lambda: run_app(main)



#!/usr/bin/env python
"""
Listing All Defined Resources

We scan ALL modules for an operations.resources.py
"""


import os, sys
from functools import partial
import importlib
from devapp.app import app, do, run_app, system
from devapp.tools import FLG, project, exists, repl_dollar_var_with_env_val

from . import _resources_api as api


class Flags(api.CommonFlags):
    pass


# ------------------------------------------------------------------------- end actions
def run():
    rscs = do(api.find_resource_defs)
    if FLG.install_state:
        do(api.add_install_state, rscs=rscs)
    l = api.to_list(rscs)
    return l


main = partial(run_app, run, flags=Flags)



#!/usr/bin/env python
"""
Simple process maintainance tools (intended for CI tests)

"""


import os, sys, json, time
from functools import partial

from devapp import app
from devapp.tools import (
    FLG,
    project,
    exists,
    jsondiff,
    write_file,
    read_file,
    offset_port,
    wait_for_port,
    repl_dollar_var_with_env_val,
)

now = time.time
app, run_app, do, system = app.app, app.run_app, app.do, app.system
from subprocess import Popen


class Flags:
    """Simple Process Start/Stop, Intended for Dev and Py-Testing"""

    autoshort = ''

    class dir:
        n = 'Change to this dir before running. env vars supported'

    class cmd:
        n = 'Process to run, with args - will be evalled'

    class background:
        s = 'bg'
        n = 'Run in background'
        d = False

    class await_port:
        n = 'Block until port is reachable, when run in background'
        d = 0

    class await_port_timeout:
        d = 10

    class kill_by_port:
        n = 'Kill matching running resource. Requires lsof'
        d = 0

    class kill_signal:
        d = 15


# --------------------------------------------------------------------------- app


fn_run = lambda: project.root() + '/tmp/running_processes.json'

run_file = lambda: json.loads(read_file(fn_run(), dflt='{}'))


def create_run_file_entry(cmd, pid):
    r = run_file()
    r[cmd] = {
        'pid': pid,
        'at': time.ctime(),
        'ts': time.time(),
    }
    # todo: maintain the file here(check pids, remove)
    write_file(fn_run(), json.dumps(r, indent=4, sort_keys=True), mkdir=True)


def run_foregrounded(cmd):
    pid = os.getpid()
    create_run_file_entry(cmd, pid)
    return do(system, cmd)


def run_backgrounded(cmd):
    app.info('Backgrounding', cmd=cmd)
    fnull = open(os.devnull, 'w')
    t0 = now()
    p = Popen(cmd, stderr=fnull, stdout=fnull, shell=True, close_fds=True)
    port = FLG.await_port
    if port:
        to = FLG.await_port_timeout
        wait_for_port(port, timeout=to)
        if now() - t0 > to:
            app.error('Port not open - killing hard', port=port, cmd=cmd)
            os.kill(p.pid, 9)
            app.die('App did not start', cmd=cmd)
        app.info('Port open', port=port, cmd=cmd)
        time.sleep(0.2)
        do(create_run_file_entry, cmd=cmd, pid=p.pid)


def kill_by_port(port):
    cmd = 'lsof -n -i :%s |grep LISTEN' % port
    r = os.popen(cmd).read().split()
    if len(r) > 2 and r[1].isdigit():
        sig = FLG.kill_signal
        app.warn('killing', port=port, pid=int(r[1]), signal=sig)
        try:
            os.kill(int(r[1]), sig)
        except Exception as ex:
            app.die('Could not kill', port=port, pid=int(r[1]), signal=signal)
    else:
        app.die('Not found open', port=port)


def run():
    if FLG.dir:
        os.chdir(repl_dollar_var_with_env_val(FLG.dir))
    p = FLG.kill_by_port
    if p:
        do(kill_by_port, port=p)
    cmd = FLG.cmd
    if cmd:
        return do(run_backgrounded if FLG.background else run_foregrounded, cmd)


#     check_running()

#     if not cmd and not FLG.killall:
#         return S.running_at_start
#     if FLG.kill or FLG.killall:
#         return kill_resource()

#     cmd, rsc = do(find_resource_by_cmd, cmd=cmd)
#     if cmd in S.running_at_start:
#         return app.warn('Already running', cmd=cmd, **S.running_at_start[cmd])

#     if FLG.show:
#         return cmd
#     if FLG.shell:
#         if FLG.background:
#             app.die('Cannot run a shell backgrounded')
#         print('Running bash instead of:\n"%s"' % cmd)
#         return os.system('bash')


main = partial(run_app, run, flags=Flags)


# def check_running():
#     S.fn_run = fn = fn_run()
#     r = json.loads(read_file(fn, dflt='{}'))
#     S.running_at_start = r = {k: v for k, v in r.items() if alive(v)}
#     write_file(fn, json.dumps(r, indent=4))


# def alive(p):
#     try:
#         return not os.kill(p['pid'], 0)
#     except:
#         return False


# def kill_resource():
#     fn, r = S.fn_run, S.running_at_start
#     for k, v in r.items():
#         if not FLG.cmd in k and not FLG.killall:
#             continue
#         app.warn('Killing resource', cmd=k, **v)
#         os.kill(v['pid'], 15)


# def find_resource_by_cmd(cmd):
#     a, args = sys.argv, []
#     orig_cmd = cmd
#     if '--' in sys.argv:
#         args = a[a.index('--') + 1 :]
#     cmd += ' ' + ' '.join(args)
#     scmd, args = cmd.split(' ', 1)
#     rsc = api.matching_resource(scmd)

#     have = [rsc.cmd] + g(rsc, 'provides', [])
#     sel = [c for c in have if scmd in c]
#     if len(sel) > 1:
#         app.info('more matches, trying exact match', matches=sel)
#         sel = [c for c in sel if scmd == c]
#     if not sel or len(sel) > 1:
#         app.die('No match found', have=have)
#     sel = sel[0]
#     breakpoint()  # FIXME BREAKPOINT
#     cmd = api.Run.find_resource_cmd(rsc, sel)
#     breakpoint()  # FIXME BREAKPOINT
#     return cmd
